import numpy as np; import pandas as pd
from pyg_timeseries._math import stdev_calculation_ewm, skew_calculation, cor_calculation_ewm, covariance_calculation, corr_calculation_ewm, LR_calculation_ewm, variance_calculation_ewm, _w
from pyg_timeseries._decorators import compiled, first_, _data_state
from pyg_timeseries._expanding import cumsum
from pyg_base import pd2np, clock, loop_all, loop, is_pd, is_df, presync, df_concat

__all__ = ['ewma', 'ewmstd', 'ewmvar', 'ewmskew', 'ewmrms',  'ewmcor',  'ewmcorr', 'ewmLR', 'ewmGLM',
           'ewma_', 'ewmstd_', 'ewmskew_', 'ewmrms_', 'ewmcor_', 'ewmvar_','ewmLR_', 'ewmGLM_',]


############################################
##
## compiled functions, unfortunately, both these methods are much slower
##
###########################################

# import numba
# from numba import int32, float32    # import the types
# from numba.experimental import jitclass

# spec = [
#     ('t0',  float32),               # a simple scalar field
#     ('t1',  float32),               # a simple scalar field
#     ('t',   float32),               # a simple scalar field
#     ('a',   float32),               # a simple scalar field
#     ('w',   float32),               # a simple scalar field
# ]

# @jitclass(spec)
# class c_ewma(object):
#     def __init__(self, a, t, t0, t1, w):
#         self.a = a
#         self.t = t
#         self.t0 = t0
#         self.t1 = t1
#         self.w = w

#     def push(self, a, t):
#         if np.isnan(a):
#             return np.nan
#         if t == self.t:
#             self.t1 = self.t1 + (1-self.w) * (a - self.a)
#             return self.t1/self.t0
#         else:
#             p = self.w if np.isnan(t) else self.w**(t-self.t)
#             self.t0 = self.t0 * p + (1-self.w)
#             self.t1 = self.t1 * p + (1-self.w) * a
#             return self.t1/self.t0


# @compiled
# def _ewma_(ai, ti, ai0, ti0, t0, t1, w):
#     """
#     we receive 
    
#     - current values, (ai, ti)
#     - previous values (ai0, ti0)
#     - current state of the moments t0, t1
#     - parameters, w
    
#     We return:
#     result, current values, updated moments
#     res, a, t, t0, t1
#     """
#     if np.isnan(ai):
#         res = np.nan
#         return res, ai0, ti0, t0, t1
#     else:
#         if ti == ti0:
#             t1 = t1 + (1-w) * (ai - ai0)
#             res = t1/t0
#             return res, ai, ti, t0, t1 
#         else:
#             p = w**(ti-ti0)
#             t0 = t0 * p + (1-w)
#             t1 = t1 * p + (1-w) * ai
#             res = t1/t0
#             return res, ai, ti, t0, t1

@pd2np
@compiled
def _ewma(a, n, time, t = np.nan, t0 = 0, t1 = 0):
    if n == 1:
        return a, t, t0, t1
    w = _w(n)
    res = np.empty_like(a)
    i0 = 0
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            if time[i] == t:
                t1 = t1 + (1-w) * (a[i] - a[i0])
            else:
                p = w if np.isnan(time[i]) else w**(time[i]-t)
                t0 = t0 * p + (1-w)
                t1 = t1 * p + (1-w) * a[i]
                t = time[i]
            i0 = i
            res[i] = np.nan if t0 == 0 else t1/t0
    return res, t, t0, t1


@pd2np
@compiled
def _ewmrms(a, n, time, t = np.nan, t0 = 0., t2 = 0., exc_zero = False, max_move = 0):
    if n == 1:
        return a, t, t0, t2
    w = _w(n)
    res = np.empty_like(a)
    apply_max = max_move>0
    i0 = 0; ai0 = a[i0]
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            if apply_max and not np.isnan(res[i0]):
                ai = max(min(res[i0] * max_move, a[i]), -res[i0] * max_move)
            else:
                ai = a[i]                
            if exc_zero and ai == 0:
                res[i] = res[i0]
            elif time[i] == t:
                t2 = t2 + (1-w) * (ai**2 - ai0**2)
            else:
                p = w if np.isnan(time[i]) else w**(time[i]-t)
                v = ai
                t0 = t0 * p + (1-w)
                t2 = t2 * p + (1-w) * v**2
                t = time[i]
            i0 = i; ai0 = ai
            res[i] = np.nan if t0 == 0 else np.sqrt(t2/t0)
    return res, t, t0, t2



@pd2np
@compiled
def _ewmstd(a, n, time, t = np.nan, t0 = 0, t1 = 0, t2 = 0, w2 = 0, min_sample = 0.25, bias = False, exc_zero = False, max_move = 0, calculator = stdev_calculation_ewm):
    if n == 1:
        return np.full_like(a, 0.0), t, t0, t1, t2, w2
    w = _w(n)
    res = np.empty_like(a)
    apply_max = max_move>0
    i0 = 0; ai0 = a[i0]
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            if apply_max and not np.isnan(res[i0]):
                ai = max(min(res[i0] * max_move, a[i]), -res[i0] * max_move)
            else:
                ai = a[i]                
            if exc_zero and ai == 0:
                res[i] = res[i0]
            elif time[i] == t:
                t1 = t1 + (1-w) * (ai - ai0)
                t2 = t2 + (1-w) * (ai**2 - ai0**2)
            else:
                p = w if np.isnan(time[i]-t) else w**(time[i]-t)
                v = ai
                t0 = t0 * p + (1-w)
                w2 = w2 * p**2 + (1-w)**2
                t1 = t1 * p + (1-w) * v
                t2 = t2 * p + (1-w) * v**2
                t = time[i]
            i0 = i; ai0 = ai
            res[i] = calculator(t0, t1, t2, w2 = w2, min_sample = min_sample, bias = bias)
    return res, t, t0, t1, t2, w2


@pd2np
@compiled
def _ewmcor(a, b, ba, n, time, t = np.nan, t0 = 0, a1 = 0, a2 = 0, b1 = 0, b2 = 0, ab = 0, w2 = 0, min_sample = 0.25, bias = False):
    """
    _ewmcor(a, b, ba, n, time, t)
    n = 50
    t = np.nan; t0 = 0; a1 = 0; a2 = 0; b1 = 0; b2 = 0; ab = 0; w2 = 0; min_sample = 0.25; bias = False    
    data, t, t0, a1, a2, b1, b2, ab, w2 = _ewmcor(a, b, ba, 200, time, t = np.nan, t0 = 0, a1 = 0, a2 = 0, b1 = 0, b2 = 0, ab = 0, w2 = 0, min_sample = 0.25, bias = False)
    pd.Series(data, drange(-9999)).plot()

    """
    if n == 1:
        return np.full_like(a, np.nan), t, t0, a1, a2, b1, b2, ab, w2
    w = _w(n)
    res = np.empty_like(a)
    i0 = 0
    for i in range(a.shape[0]):
        if np.isnan(a[i]) or np.isnan(b[i]):
            res[i] = np.nan
        else:
            if time[i] == t:
                a1 = a1 + (1-w) * (a[i] - a[i0])
                a2 = a2 + (1-w) * (a[i]**2 - a[i0]**2)
                b1 = b1 + (1-w) * (b[i] - b[i0])
                b2 = b2 + (1-w) * (b[i]**2 - b[i0]**2)
                ab = ab + (1-w) * (ba[i] - ba[i0])
            else:
                p = w if np.isnan(time[i]) else w**(time[i]-t)
                t0 = t0 * p + (1-w)
                w2 = w2 * p**2 + (1-w)**2
                a1 = a1 * p + (1-w) * a[i]
                a2 = a2 * p + (1-w) * a[i]**2
                b1 = b1 * p + (1-w) * b[i]
                b2 = b2 * p + (1-w) * b[i]**2
                ab = ab * p + (1-w) * ba[i]
                t = time[i]
            i0 = i
            res[i0] = cor_calculation_ewm(t0 = t0, a1 = a1, a2 = a2, b1 = b1, b2 = b2, ab = ab, w2 = w2, min_sample = min_sample, bias = bias)
    return res, t, t0, a1, a2, b1, b2, ab, w2


@compiled
def _ewmcorr(a, n, a0 = None, a1 = None, a2 = None, aa0 = None, aa1 = None, w2 = None, min_sample = 0.25, bias = False):
    """

    """
    m = a.shape[1]
    if n == 1:
        return np.full((a.shape[0], m, m), np.nan), a0, a1, a2, aa0, aa1, w2
    p = w = _w(n)
    v = 1 - w
    res = np.zeros((a.shape[0], m, m))
    a0 = np.zeros(m) if a0 is None else a0
    a1 = np.zeros(m) if a1 is None else a1
    a2 = np.zeros(m) if a2 is None else a2
    aa1 = np.zeros((m,m)) if aa1 is None else aa1
    aa0 = np.zeros((m,m)) if aa0 is None else aa0
    w2 = np.zeros(m) if w2 is None else w2
    for i in range(a.shape[0]):
        for j in range(m):
            if ~np.isnan(a[i,j]):
                w2[j] = w2[j] * p**2 + v**2
                a0[j] = a0[j] * p + v
                a1[j] = a1[j] * p + v * a[i,j]
                a2[j] = a2[j] * p + v * a[i,j] ** 2        
        for j in range(m):
            res[i, j, j] = 1.
            if np.isnan(a[i,j]):
                res[i, j, :] = np.nan #if i == 0 else res[i-1, j, :] # we ffill correlations
                res[i, :, j] = np.nan #if i == 0 else res[i-1, :, j]
            else:
                for k in range(j):
                    if ~np.isnan(a[i,k]):
                        aa0[j,k] = aa0[j,k] * p + v 
                        aa1[j,k] = aa1[j,k] * p + v * a[i, j] * a[i, k]
                        res[i, k, j] = res[i, j, k] = corr_calculation_ewm(a0 = a0[j], a1 = a1[j], a2 = a2[j], aw2 = w2[j], 
                                                                           b0 = a0[k], b1 = a1[k], b2 = a2[k], bw2 = w2[k],
                                                                           ab = aa1[j,k], ab0 = aa0[j,k],
                                                                           min_sample = min_sample, bias = bias)                    
    return res, a0, a1, a2, aa0, aa1, w2



@compiled
def _ewmcorrelation(a, n, a0 = None, a1 = None, a2 = None, aa = None, prev = None, w2 = None, min_sample = 0.25, bias = False):
    """
    synchronized correlation calculation
    a is assumed to be total returns as opposed to deltas
    
    For each pair j>k we will store
    prev[j,k] : prev value of j when j,k were non-nan
    prev[k,j] : prev value of k when j,k were non-nan

    a0[j,k] : the total 0th moment of j on the pairing (j,k)
    a1[j,k] : the total 1st moment of j on the pairing (j,k)
    a2[j,k] : the total 2nd moment of j on the pairing (j,k)

    a0[j,k] : the total 0th moment of k on the pairing (j,k)
    a1[j,k] : the total 1st moment of k on the pairing (j,k)
    a2[j,k] : the total 2nd moment of k on the pairing (j,k)
    
    """
    m = a.shape[1]
    if n == 1:
        return np.full((a.shape[0], m, m), np.nan), a0, a1, a2, aa, prev, w2
    p = w = _w(n)
    v = 1 - w
    res = np.full((a.shape[0], m, m), np.nan)
    prev = np.full((m,m), np.nan) if prev is None else prev
    a0 = np.zeros((m,m)) if a0 is None else a0
    a1 = np.zeros((m,m)) if a1 is None else a1
    a2 = np.zeros((m,m)) if a2 is None else a2
    aa = np.zeros((m,m)) if aa is None else aa
    w2 = np.zeros((m,m)) if w2 is None else w2
    for i in range(a.shape[0]):
        for j in range(m):
            res[i, j, j] = 1.
        for j in range(m):
            if ~np.isnan(a[i,j]):
                for k in range(j):
                    if ~np.isnan(a[i,k]):
                        if ~np.isnan(prev[j,k]) and ~np.isnan(prev[k,j]):
                            dx = a[i,j] - prev[j,k]
                            dy = a[i,k] - prev[k,j]
                            w2[j,k] = w2[k,j] = w2[j,k] * p**2 + v**2       
                            a0[j,k] = a0[k,j] = a0[j,k] * p + v
                            a1[j,k] = a1[j,k] * p + v * dx
                            a1[k,j] = a1[k,j] * p + v * dy
                            a2[j,k] = a2[j,k] * p + v * dx ** 2        
                            a2[k,j] = a2[k,j] * p + v * dy ** 2                                
                            aa[j,k] = aa[k,j] = aa[j,k] * p + v * dx * dy
                            res[i, k, j] = res[i, j, k] = corr_calculation_ewm(a0 = a0[j,k], a1 = a1[j,k], a2 = a2[j,k], aw2 = w2[j,k], 
                                                                               b0 = a0[k,j], b1 = a1[k,j], b2 = a2[k,j], bw2 = w2[k,j],
                                                                               ab = aa[j,k], ab0 = a0[j,k],
                                                                               min_sample = min_sample, bias = bias)
                        prev[j,k] = a[i,j]
                        prev[k,j] = a[i,k]
                            
    return res, a0, a1, a2, aa, prev, w2



@compiled
def _ewmcovar(a, n, a0 = None, a1 = None, aa0 = None, aa1 = None, min_sample = 0.25, bias = False):
    """

    """
    m = a.shape[1]
    if n == 1:
        return np.full((a.shape[0], m, m), np.nan), a0, a1, aa0, aa1
    p = w = _w(n)
    v = 1 - w
    res = np.zeros((a.shape[0], m, m))
    a0 = np.zeros(m) if a0 is None else a0
    a1 = np.zeros(m) if a1 is None else a1
    aa1 = np.zeros((m,m)) if aa1 is None else aa1
    aa0 = np.zeros((m,m)) if aa0 is None else aa0
    for i in range(a.shape[0]):
        for j in range(m):
            if ~np.isnan(a[i,j]):
                a0[j] = a0[j] * p + v
                a1[j] = a1[j] * p + v * a[i,j]
        for j in range(m):
            if np.isnan(a[i,j]):
                res[i, j, :] = np.nan #if i == 0 else res[i-1, j, :] # we ffill correlations
                res[i, :, j] = np.nan #if i == 0 else res[i-1, :, j]
            else:
                for k in range(j+1):
                    if ~np.isnan(a[i,k]):
                        aa0[j,k] = aa0[j,k] * p + v 
                        aa1[j,k] = aa1[j,k] * p + v * a[i, j] * a[i, k]
                        res[i, k, j] = res[i, j, k] = covariance_calculation(a0 = a0[j], a1 = a1[j], 
                                                                           b0 = a0[k], b1 = a1[k],
                                                                           ab = aa1[j,k], ab0 = aa0[j,k],
                                                                           min_sample = min_sample, bias = bias)                    
    return res, a0, a1, aa0, aa1

@compiled
def _ewmcovariance(a, n, a0 = None, a1 = None, aa = None, prev = None, min_sample = 0.25, bias = False):
    """

    """
    m = a.shape[1]
    if n == 1:
        return np.full((a.shape[0], m, m), np.nan), a0, a1, aa, prev
    p = w = _w(n)
    v = 1 - w
    res = np.full((a.shape[0], m, m), np.nan)
    prev = np.full((m,m), np.nan) if prev is None else prev
    a0 = np.zeros((m,m)) if a0 is None else a0
    a1 = np.zeros((m,m)) if a1 is None else a1
    aa = np.zeros((m,m)) if aa is None else aa
    for i in range(a.shape[0]):
        for j in range(m):
            if ~np.isnan(a[i,j]):
                for k in range(j+1):
                    if ~np.isnan(a[i,k]):
                        if ~np.isnan(prev[j,k]) and ~np.isnan(prev[k,j]):
                            dx = a[i,j] - prev[j,k]
                            dy = a[i,k] - prev[k,j]
                            a0[j,k] = a0[k,j] = a0[j,k] * p + v
                            a1[j,k] = a1[j,k] * p + v * dx
                            a1[k,j] = a1[k,j] * p + v * dy
                            aa[j,k] = aa[k,j] = aa[j,k] * p + v * dx * dy
                            res[i, k, j] = res[i, j, k] = covariance_calculation(a0 = a0[j,k], a1 = a1[j,k], 
                                                                           b0 = a0[k,j], b1 = a1[k,j],
                                                                           ab = aa[j,k], ab0 = a0[j,k],
                                                                           min_sample = min_sample, bias = bias)                    
                        prev[j,k] = a[i,j]
                        prev[k,j] = a[i,k]
    return res, a0, a1, aa, prev



def ewmcovariance_(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full correlation matrix as a timeseries. Also returns the recent state of the calculations.
    
    :Returns:
    ---------
        a dict with:
            - data: t x n x n covariance matrix
            - index: timeseries index
            - columns: columns of original data

    See ewmcorr for full details.
    
    """
    state = {} if instate is None else instate
    arr = df_concat(a, join = join, method = method)
    if isinstance(arr, np.ndarray):
        res, a0, a1, aa, prev = _ewmcovariance(arr, n, min_sample = min_sample, bias = bias, **state)
        # if res.shape[1] == 2:
        #     res = res[:, 0, 1]
        return dict(data = res, columns = None, index = None, state = dict(a0=a0, a1=a1, aa=aa, prev = prev))
    elif is_df(arr):
        index = arr.index
        columns = list(arr.columns)
        res, a0, a1, aa, prev = _ewmcovariance(arr.values, n, min_sample = min_sample, bias = bias, **state)
        state = dict(a0=a0, a1=a1, aa=aa, prev = prev)
        return dict(data = res, columns = columns, index = index, state = state)
    else:
        raise ValueError('unsure how to calculate correlation matrix for a %s'%a)

ewmcovariance_.output = ['data', 'columns', 'index', 'state']


def ewmcovariance(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full covariance matrix as a timeseries. 

    :Parameters:
    ----------
    a : np.array or a pd.DataFrame
        multi-variable timeseries to calculate correlation for
    n : int
        days for which rolling correlation is calculated.
    min_sample : float, optional
        Minimum observations needed before we calculate correlation. The default is 0.25.
    bias : bool, optional
        input to stdev calculations, the default is False.
    instate : dict, optional
        historical calculations so far.

    :Returns:
    -------
    covariance (as t x n x n np.array)
        
        
    :Example: a pair of ts
    ---------
    >>> a = pd.DataFrame(np.random.normal(0,1,(10000,10)), drange(-9999))
    >>> cov1_ = ewmcovariance_(cumsum(a), 250)
    >>> cov2_ = ewmcovar_(a, 250)
    >>> cov1 = cov1_['data']
    >>> cov2 = cov2_['data']
    >>> vol1 = pd.DataFrame([[row[j,j] for j in range(10)] for row in cov1], a.index) ** 0.5
    >>> vol2 = pd.DataFrame([[row[j,j] for j in range(10)] for row in cov2], a.index) ** 0.5
    >>> vol1.plot()
    >>> vol2.plot()

    
    >>> # We first check that diagonal is indeed the (biased) variance of the variables: 
    >>> ratio = pd.Series([cov1[-1][i,i] for i in range(10)]) / ewmvar(a, 250, bias=True).iloc[-1]
    >>> assert ratio.max() < 1.0001 and ratio.min() > 0.9999
    
    To access individually, here we calculate the correlation between 0th and 1st timeseries. That correlation is close to 0 (Fisher distribution) so...
    
    >>> cor = pd.Series(cov1[:,0,1] / np.sqrt(cov1[:,0,0] * cov1[:,1,1]), a.index)
    >>> cor.plot()
    >>> assert cor.max() < 0.3 and cor.min() > -0.3
    """
    return ewmcovariance_(a, n, min_sample = min_sample, bias = bias, instate = instate , join = join, method = method).get('data')



def ewmcovar_(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full covariance matrix as a timeseries. Also returns the recent state of the calculations.
    
    :Returns:
    ---------
        a dict with:
            - data: t x n x n covariance matrix
            - index: timeseries index
            - columns: columns of original data

    See ewmcovar for full details.
    
    """
    state = {} if instate is None else instate
    arr = df_concat(a, join = join, method = method)
    state['prev'] = state.pop('prev', np.zeros((arr.shape[1], arr.shape[1])))
    return ewmcovariance_(cumsum(arr), n = n, min_sample=min_sample, bias = bias, instate = state, join = join, method = method)   

ewmcovar_.output = ['data', 'columns', 'index', 'state']


def ewmcovar(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full covariance matrix as a timeseries. 

    :Parameters:
    ----------
    a : np.array or a pd.DataFrame
        multi-variable timeseries to calculate correlation for
    n : int
        days for which rolling correlation is calculated.
    min_sample : float, optional
        Minimum observations needed before we calculate correlation. The default is 0.25.
    bias : bool, optional
        input to stdev calculations, the default is False.
    instate : dict, optional
        historical calculations so far.

    :Returns:
    -------
    covariance (as t x n x n np.array)
        
        
    :Example: a pair of ts
    ---------
    >>> a = pd.DataFrame(np.random.normal(0,1,(10000,10)), drange(-9999))
    >>> res = ewmcovar(a, 250)
    
    >>> # We first check that diagonal is indeed the (biased) variance of the variables: 
    >>> ratio = pd.Series([res[-1][i,i] for i in range(10)]) / ewmvar(a, 250, bias=True).iloc[-1]
    >>> assert ratio.max() < 1.02 and ratio.min() > 0.98
    
    To access individually, here we calculate the correlation between 0th and 1st timeseries. That correlation is close to 0 (Fisher distribution) so...
    
    >>> cov = pd.Series(res[:,0,1] / np.sqrt(res[:,0,0] * res[:,1,1]), a.index)
    >>> cov.plot()
    >>> assert cov.max() < 0.3 and cor.min() > -0.3
    """
    return ewmcovar_(a, n, min_sample = min_sample, bias = bias, instate = instate , join = join, method = method).get('data')


def ewmcorr_(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full correlation matrix as a timeseries. Also returns the recent state of the calculations.
    See ewmcorr for full details.
    
    """
    state = {} if instate is None else instate
    arr = df_concat(a, join = join, method = method)
    state['prev'] = state.pop('prev', np.zeros((arr.shape[1], arr.shape[1])))
    return ewmcorrelation_(cumsum(arr), n = n, min_sample=min_sample, bias = bias, instate = state, join = join, method = method)   

ewmcorr_.output = ['data', 'state', 'index', 'columns']


def ewmcorr(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full correlation matrix as a timeseries. 

    :Parameters:
    ----------
    a : np.array or a pd.DataFrame
        multi-variable timeseries to calculate correlation for
    n : int
        days for which rolling correlation is calculated.
    min_sample : float, optional
        Minimum observations needed before we calculate correlation. The default is 0.25.
    bias : bool, optional
        input to stdev calculations, the default is False.
    instate : dict, optional
        historical calculations so far.

    :Returns:
    -------
    correlation dataset
        an xarray.Dataset unless there are only two timeseries in a, in which case, a single column correlation timeseries is returned
        
        
    :Example: a pair of ts
    ---------
    >>> rtn = np.random.normal(0,1,10000)
    >>> adj = cumsum(rtn)
    >>> x0 = ewmacd(adj, 10, 20, 30)[50:]
    >>> x1 = ewmacd(adj, 20, 40, 30)[50:]
    >>> a = pd.DataFrame(np.array([x0,x1]).T, drange(-9949))
    >>> res = ewmcorr(a, n)
    >>> ts = pd.Series(res[:,0,1], a.index)

    >>> ts    
    >>>     Out[130]: 
    >>>     1994-06-07         NaN
    >>>     1994-06-08         NaN
    >>>     1994-06-09         NaN
    >>>     1994-06-10         NaN
    >>>     1994-06-11         NaN
      
    >>>     2021-08-29    0.890766
    >>>     2021-08-30    0.886926
    >>>     2021-08-31    0.883054
    >>>     2021-09-01    0.879577
    >>>     2021-09-02    0.875766
    >>>     Length: 9950, dtype: float64    

    :Example: multi column ts
    ---------
    >>> rtn = np.random.normal(0,1,10000)
    >>> adj = cumsum(rtn)
    >>> x0 = ewmacd(adj, 10, 20, 30)[50:]
    >>> x1 = ewmacd(adj, 20, 40, 30)[50:]
    >>> x2 = ewmacd(adj, 40, 80, 30)[50:]
    >>> a = pd.DataFrame(np.array([x0,x1,x2]).T, drange(-9949), ['a','b','c'])
    >>> ds = ewmcorr_(a, n)
    >>> ds.keys()
    >>> dict_keys(['data', 'columns', 'index', 'state'])
    
    >>> ds['data'].shape
    >>> (9950, 3, 3)

    To access individual correlations:
    
    >>> a_vs_b = pd.Series(ds['data'][:, 1, 2], ds['index'])
    
    To access all correlations to a:    
    
    >>> a_vs_all = pd.DataFrame(ds['data'][:,:,0], ds['index'], ds['columns'])
    >>> a_vs_all.plot()
    >>> a_vs_all
    
                  a         b         c
    1997-12-13  1.0       NaN       NaN
    1997-12-14  1.0       NaN       NaN
    1997-12-15  1.0       NaN       NaN
    1997-12-16  1.0       NaN       NaN
    1997-12-17  1.0       NaN       NaN
            ...       ...       ...
    2025-03-06  1.0  0.895512  0.778449
    2025-03-07  1.0  0.884339  0.753600
    2025-03-08  1.0  0.871431  0.726200
    2025-03-09  1.0  0.858903  0.699254
    2025-03-10  1.0  0.846466  0.673574
    
    """
    return ewmcorr_(a, n, min_sample = min_sample, bias = bias, instate = instate , join = join, method = method).get('data')

def ewmcorrelation_(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full correlation matrix as a timeseries. Also returns the recent state of the calculations.
    See ewmcorrelation for full details.
    
    """
    state = {} if instate is None else instate
    arr = df_concat(a, join = join, method = method)
    if isinstance(arr, np.ndarray):
        res, a0, a1, a2, aa, prev, w2 = _ewmcorrelation(arr, n, min_sample = min_sample, bias = bias, **state)
        return dict(data = res, index = None, columns = None, state = dict(a0=a0, a1=a1, a2=a2, aa=aa, prev = prev, w2 = w2))
    elif is_df(arr):
        index = arr.index
        columns = list(arr.columns)
        res, a0, a1, a2, aa, prev, w2 = _ewmcorrelation(arr.values, n, min_sample = min_sample, bias = bias, **state)
        state = dict(a0=a0, a1=a1, a2=a2, aa=aa, prev=prev, w2 = w2)
        return dict(data = res, columns = columns, index = index, state = state)
    else:
        raise ValueError('unsure how to calculate correlation matrix for a %s'%a)

ewmcorrelation_.output = ['data', 'state', 'index', 'columns']


def ewmcorrelation(a, n, min_sample = 0.25, bias = False, instate = None, join = 'outer', method = None):
    """
    This calculates a full correlation matrix as a timeseries. 

    :Parameters:
    ----------
    a : np.array or a pd.DataFrame
        multi-variable timeseries of aggregate returns to calculate correlation for
    n : int
        days for which rolling correlation is calculated.
    min_sample : float, optional
        Minimum observations needed before we calculate correlation. The default is 0.25.
    bias : bool, optional
        input to stdev calculations, the default is False.
    instate : dict, optional
        historical calculations so far.

    :Returns:
    -------
    correlation dataset
        an xarray.Dataset unless there are only two timeseries in a, in which case, a single column correlation timeseries is returned
        
        
    :Example: multi column ts
    ---------
    >>> join = 'outer'; method = None
    >>> rtn = np.random.normal(0,1,10000)
    >>> adj = cumsum(rtn)
    >>> x0 = ewmacd(adj, 10, 20, vol = 18)
    >>> x1 = ewmacd(adj, 20, 40, vol = 18)
    >>> x2 = ewmacd(adj, 40, 80, vol = 18)
    >>> x0[np.random.normal(0,1,10000)>0] = np.nan
    >>> x1[np.random.normal(0,1,10000)>0] = np.nan
    >>> x2[np.random.normal(0,1,10000)>0] = np.nan
    >>> a = pd.DataFrame(np.array([x0,x1,x2]).T, drange(-9999), ['a','b','c'])
    >>> ds = ewmcorrelation(a, 30)
    >>> c = ewmcorr(diff(a), 30)
    >>> ds.shape
    >>> (10000, 3, 3)

    To access individual correlations:
    
    >>> a_vs_b = pd.Series(ds[:, 1, 2], a.index)
    >>> a_vs_b.plot() 
    
    
    To access all correlations to a:
    >>> a_vs_all = pd.DataFrame(ds[:,:,0], a.index, a.columns)
    >>> a_vs_all.ffill().plot()
        
    """
    return ewmcorrelation_(a, n, min_sample = min_sample, bias = bias, instate = instate , join = join, method = method).get('data')


@pd2np
@compiled
def _ewmLR(a, b, ba, n, time, t = np.nan, t0 = 0, a1 = 0, a2 = 0, b1 = 0, b2 = 0, ab = 0, w2 = 0, min_sample = 0.25, bias = False):
    """
    We have a and b for which we want to fit:
        b_i = c + m a_i 
    
    minimize LSE under weights w_i. We let:
        w = \sum_i w_i
    
    LSE(c,m) = \sum w_i (c + m a_i - b_i)^2
    dLSE/dc  = 0  <==> \sum w_i  (c + m a_i - b_i) = 0    [1]
    dLSE/dm  = 0 <==> \sum w_i  a_i (c + m a_i - b_i) = 0 [2]

    c     + mE(a)    = E(b)     [1]
    cE(a) + mE(a^2)  = E(ab)    [2]
    
    cE(a) + mE(a)^2  = E(a)E(n) [1] * E(a) 
    m(E(a^2) - E(a)^2) = E(ab) - E(a)E(b)
    
    m = covar(a,b)/var(a)
    c = E(b) - mE(a)
    

    :Example:
    -----------
    a = np.random.normal(0,1,10000)
    b = 0.5 * a + np.random.normal(0,1,10000)
    _ewmLRt(a,b,100)
    """
    w = _w(n)
    m = np.empty_like(a)
    c = np.empty_like(a)
    i0 = 0
    for i in range(a.shape[0]):
        if np.isnan(a[i]) or np.isnan(b[i]):
            c[i] = m[i] = np.nan
        else:
            if time[i] == t:
                a1 = a1 + (1-w) * (a[i] - a[i0])
                a2 = a2 + (1-w) * (a[i]**2 - a[i0]**2)
                b1 = b1 + (1-w) * (b[i] - b[i0])
                b2 = b2 + (1-w) * (b[i]**2 - b[i0]**2)
                ab = ab + (1-w) * (ba[i] - ba[i0])
            else:
                p = w if np.isnan(time[i]) else w**(time[i]-t)
                t0 = t0 * p + (1-w)
                w2 = w2 * p**2 + (1-w)**2
                a1 = a1 * p + (1-w) * a[i]
                a2 = a2 * p + (1-w) * a[i]**2
                b1 = b1 * p + (1-w) * b[i]
                b2 = b2 * p + (1-w) * b[i]**2
                ab = ab * p + (1-w) * ba[i]
                t = time[i]
            i0 = i
            c[i0], m[i0] = LR_calculation_ewm(t0 = t0, a1 = a1, a2 = a2, b1 = b1, b2 = b2, ab = ab, w2 = w2, min_sample = min_sample, bias = bias)
    return c, m, t, t0, a1, a2, b1, b2, ab, w2


@compiled
def xTx(x):
    res = np.empty((x.shape[0], x.shape[0]))
    for i in range(x.shape[0]):
        for j in range(i):
            res[i][j] = res[j][i] = x[i] * x[j]
        res[i][i] = x[i]**2
    return res


@compiled
def _ewmGLM(a, b, n, time, t, t0, a2, ab, min_weight = 0.25):
    """
    We assume b is single column while a is multicolumn. We are fitting
    
    b[i] =\sum_j m_j a_j[i]
    
    LSE(m) = \sum_i w_i (b[i] - \sum_j m_j * a_j[i])^2
    dLSE/dm_k = 0  
    <==>  \sum_i w_i (b[i] - \sum_j m_j * a_j[i]) a_k[i] = 0
    <==>  E(b*a_k) = m_k E(a_k^2) + sum_{j<>k} m_k E(a_j a_k) 
    
    E is expectation under w. 
    
    We set
    a2[i,j] = E(a_i a_j)
    ab[j] = E(a_j * b)
    
    And invert
    A*m = B
    
    Example
    --------
    >>> from pyg import *
    >>> a = np.random.normal(0,1,(5000,10))
    >>> m = np.random.normal(1,1,10)
    >>> a[a>2] = np.nan
    >>> b = np.sum(a * m,axis=1) + np.random.normal(0,1,5000)
    >>> n = 30
    >>> t = np.nan; t0 = 0; time = np.full_like(b, np.nan); a2 = np.zeros((a.shape[1],a.shape[1])); ab = np.zeros(a.shape[1])
    >>> res = _ewmGLM(a, b, n, time, t, t0, a2, ab)
    >>> pd.DataFrame(res[0], drange(-4999)).plot()
    >>> pd.Series(m)
    """
    w = _w(n)
    res = np.empty_like(a)
    nana = np.sum(np.isnan(a), axis = 1)
    i0 = 0
    for i in range(a.shape[0]):
        if nana[i]>0 or np.isnan(b[i]):
            res[i] = np.nan
        else:
            if time[i] == t:
                ab = ab + (1-w) * (b[i]*a[i] - b[i0]*a[i0])
                a2 = a2 + (1-w) * xTx(a[i]) - (1-w) * xTx(a[i0])
            else:
                p = w if np.isnan(time[i]) else w**(time[i]-t)
                t0 = t0 * p + (1-w)
                ab = ab * p + (1-w) * b[i]*a[i]
                a2 = a2 * p + (1-w) * xTx(a[i])
                t = time[i]                
            i0 = i
            if t0>min_weight:
                a2i = np.linalg.inv(a2/t0)
                res[i] = a2i.dot(ab/t0)
            else:
                res[i] = np.nan
    return res, t, t0, a2, ab
    

@pd2np
@compiled
def _ewmskew(a, n, time, bias = 0, t = np.nan, t0 = 0, t1 = 0, t2 = 0, t3 = 0, min_sample = 0.25):
    w = _w(n)
    days = n if n>1 else w/(1-w)
    d = 1 + days
    res = np.empty_like(a)
    i0 = 0
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            if time[i] == t:
                t1 = t1 + (1-w) * (a[i] - a[i0])
                t2 = t2 + (1-w) * (a[i]**2 - a[i0]**2)
                t3 = t3 + (1-w) * (a[i]**3 - a[i0]**3)
            else:
                p = w if np.isnan(time[i]) else w**(time[i]-t)
                v = a[i]
                t0 = t0 * p + (1-w)
                t1 = t1 * p + (1-w) * v
                t2 = t2 * p + (1-w) * v**2
                t3 = t3 * p + (1-w) * v**3
                t = time[i]
            i0 = i
            res[i] = skew_calculation(t0 * d, t1 * d, t2 * d, t3 * d, bias, min_sample)
    return res, t, t0, t1, t2, t3

##### clock management functions  ############################################

@loop_all
def _ewmat(a, n, time = None, t = None, t0 = 0, t1 = 0):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ewma(a = a, n = n, time = time, t = t, t0 = t0, t1 = t1)

@loop_all
def _ewmrmst(a, n, time = None, t = None, t0 = 0, t2 = 0, exc_zero = False, max_move = 0):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ewmrms(a, n, time = time, t = t, t0 = t0, t2 = t2, exc_zero = exc_zero, max_move = max_move)

@loop_all
def _ewmstdt(a, n, time = None, t = None, t0 = 0, t1 = 0, t2 = 0, w2 = 0, min_sample = 0.25, bias = False, exc_zero = False, max_move = 0, calculator = stdev_calculation_ewm):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ewmstd(a, n, time = time, t = t, t0 = t0, t1 = t1, t2 = t2, w2 = w2, min_sample=min_sample, bias = bias, calculator = calculator, exc_zero = exc_zero, max_move = max_move)



@loop_all
@presync
def _ewmcort(a, b, n, time = None, t = None, t0 = 0, a1 = 0, a2 = 0, b1 = 0, b2 = 0, ab = 0, w2 = 0, min_sample = 0.25, bias = False):
    ba = b * a
    time = clock(ba, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ewmcor(a = a, b = b, ba = ba, n = n, time = time, t = t, t0 = t0, a1 = a1, a2 = a2, b1 = b1, b2 = b2, ab = ab, w2 = w2, min_sample=min_sample, bias = bias)

@loop_all
@presync
def _ewmLRt(a, b, n, time = None, t = None, t0 = 0, a1 = 0, a2 = 0, b1 = 0, b2 = 0, ab = 0, w2 = 0, min_sample = 0.25, bias = False):
    ba = b * a
    time = clock(ba, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ewmLR(a = a, b = b, ba = ba, n = n, time = time, t = t, t0 = t0, a1 = a1, a2 = a2, b1 = b1, b2 = b2, ab = ab, w2 = w2, min_sample=min_sample, bias = bias)

@loop(list, dict)
@presync(columns = False)
def _ewmGLMt(a, b, n, time = None, t = None, t0 = 0, a2 = None, ab = None, min_sample = 0.25):
    """
    >>> from pyg import *
    >>> a = np.random.normal(0,1,(5000,10))
    >>> m = np.random.normal(1,1,10)
    >>> b = np.sum(a * m,axis=1) + np.random.normal(0,1,5000)
    >>> a[a>2] = np.nan
    >>> n = 30
    >>> time = None; t = None; t0 = 0; a2 = None; ab = None; min_sample = 0.25
    >>> res = ewmGLM_(a, b, n)
    
    """
    time = clock(b, time, t)
    t = 0 if t is None or np.isnan(t) else t
    a2 = np.zeros((a.shape[1], a.shape[1])) if a2 is None else a2
    ab = np.zeros(a.shape[1]) if ab is None else ab
    a_ = a.values if is_pd(a) else a  #np.array(list(a.values)) 
    b_ = b.values if is_pd(b) else b
    res = _ewmGLM(a = a_, b = b_, n = n, time = time, t = t, t0 = t0, a2 = a2, ab = ab, min_weight = min_sample)
    if is_pd(a):
         res = (pd.DataFrame(res[0], a.index, columns = a.columns),) + res[1:]
    return res


@loop_all
def _ewmskewt(a, n, time = None, t = None, bias = False, t0 = 0, t1 = 0, t2 = 0, t3 = 0, min_sample = 0.25):
    time = clock(a, time, t)
    t = 0 if t is None or np.isnan(t) else t
    return _ewmskew(a, n, time = time, bias = int(bias), t = t, t0 = t0, t1 = t1, t2 = t2, t3 = t3, min_sample=min_sample)    


#####   exposed API functions ############################################

def ewma_(a, n, time = None, data = None, instate = None):
    """
    Equivalent to ewma but returns a state parameter for instantiation of later calculations.
    See ewma documentation for more details
    """
    state = instate or {}
    return _data_state(['data', 't', 't0', 't1'], _ewmat(a = a, n = n, time = time, **state))

ewma_.output = ['data', 'state']


def ewma(a, n, time = None, axis = 0, data = None, state = None):
    """
    ewma is equivalent to a.ewm(n).mean() but with...
    - supports np.ndarrays as well as timeseries
    - handles nan by skipping them
    - allows state-management
    - ability to supply a 'clock' to the calculation
    
    :Parameters:
    ------------
    a : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    
    :Example: matching pandas
    -------------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> ts = ewma(a,10); df = a.ewm(10).mean()
    >>> assert abs(ts-df).max()<1e-10

    :Example: numpy arrays support
    ------------------------------
    >>> assert eq(ewma(a.values, 10), ewma(a,10).values)

    :Example: nan handling
    ----------------------
    >>> a[a.values<0.1] = np.nan
    >>> ts = ewma(a,10, time = 'i'); df = a.ewm(10).mean() # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10
    
    >>> pd.concat([ts,df], axis=1)
    >>>                        0         1
    >>> 1993-09-24  0.263875  0.263875
    >>> 1993-09-25       NaN  0.263875
    >>> 1993-09-26       NaN  0.263875
    >>> 1993-09-27       NaN  0.263875
    >>> 1993-09-28       NaN  0.263875
    >>>                  ...       ...
    >>> 2021-02-04       NaN  0.786506
    >>> 2021-02-05  0.928817  0.928817
    >>> 2021-02-06       NaN  0.928817
    >>> 2021-02-07  0.839168  0.839168
    >>> 2021-02-08  0.831109  0.831109

    :Example: state management
    --------------------------
    >>> old = a.iloc[:5000]
    >>> new = a.iloc[5000:]
    >>> old_ts = ewma_(old, 10)
    >>> new_ts = ewma(new, 10, **old_ts) # instantiation with previous ewma
    >>> ts = ewma(a,10)
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: Support for time & clock
    ------------------------------------
    >>> daily = a
    >>> monthly = daily.resample('M').last()
    >>> m_ts = ewma(monthly, 3) ## 3-month ewma run on monthly data
    >>> d_ts = ewma(daily, 3, 'm') ## 3-month ewma run on daily data
    >>> daily_resampled_to_month = d_ts.resample('M').last()
    >>> assert abs(daily_resampled_to_month - m_ts).max() < 1e-10

    So you can run a 3-monthly ewma on daily, where within month, most recent value is used with the EOM history.
    
    :Example: Support for dict/list of arrays
    --------------------------------------
    >>> x = pd.Series(np.random.normal(0,1,1000), drange(-999)); y = pd.Series(np.random.normal(0,1,1000), drange(-999))
    >>> a = dict(x = x, y = y)    
    >>> assert eq(ewma(dict(x=x, y=y),10), dict(x=ewma(x,10), y=ewma(y,10)))
    >>> assert eq(ewma([x,y],10), [ewma(x,10), ewma(y,10)])
    
             
    :Returns:
    -------
    an array/timeseries of ewma
    
    """
    state = state or {}
    return first_(_ewmat(a=a, n=n, time = time, axis=axis, **state))


def ewmrms_(a, n, time = None, axis = 0, data = None, exc_zero = False, max_move = 0, instate = None):
    """
    Equivalent to ewmrms but returns a state parameter for instantiation of later calculations.
    See ewmrms documentation for more details
    """
    state = instate or {}
    return _data_state(['data', 't', 't0', 't2'],_ewmrmst(a, n, time = time, axis=axis, exc_zero = exc_zero, max_move = max_move,**state))

ewmrms_.output = ['data', 'state']

def ewmrms(a, n, time=None, axis=0, data = None, exc_zero = False, max_move = 0, state = None):
    """
    ewmrms is equivalent to (a**2).ewm(n).mean()**0.5  but with...
    - supports np.ndarrays as well as timeseries
    - handles nan by skipping them
    - allows state-management
    - ability to supply a 'clock' to the calculation
    
    :Parameters:
    ------------
    a : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    
    exc_zero: bool
        if True, will skip zeros inside a and will assume these are a result of forward filling
    max_move: float
        if positive, will "clip" a[i] at current stdev estimation * max_move
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    
    :Example: matching pandas
    -------------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> ts = ewmrms(a,10); df = (a**2).ewm(10).mean()**0.5
    >>> assert abs(ts-df).max()<1e-10

    :Example: numpy arrays support
    ------------------------------
    >>> assert eq(ewmrms(a.values, 10), ewmrms(a,10).values)

    :Example: nan handling
    ----------------------
    >>> a[a.values<0.1] = np.nan
    >>> ts = ewmrms(a,10, time = 'i'); df = (a**2).ewm(10).mean()**0.5 # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10
    
    >>> pd.concat([ts,df], axis=1)
    >>>                        0         1
    >>> 1993-09-24  0.263875  0.263875
    >>> 1993-09-25       NaN  0.263875
    >>> 1993-09-26       NaN  0.263875
    >>> 1993-09-27       NaN  0.263875
    >>> 1993-09-28       NaN  0.263875
    >>>                  ...       ...
    >>> 2021-02-04       NaN  0.786506
    >>> 2021-02-05  0.928817  0.928817
    >>> 2021-02-06       NaN  0.928817
    >>> 2021-02-07  0.839168  0.839168
    >>> 2021-02-08  0.831109  0.831109
        
    :Example: exc_zero
    ------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> a[dt(-10):] = 0
    >>> exclude = ewmrms(a, 10, exc_zero = True)
    >>> assert exclude .iloc[-1] == exclude.iloc[-10]
    >>> no_exclude = ewmrms(a, 10)
    >>> assert no_exclude.iloc[-1] < no_exclude.iloc[-10]

    :Example: max_move
    ------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)) ## a series with stdev of 1 approximately
    >>> a.iloc[-1] = 100. ## put a bad data point with a huge spike...
    >>> max_move = ewmrms(a, 10, max_move = 4)
    >>> assert max_move.iloc[-1] < 2
    >>> no_max_move = ewmrms(a, 10)
    >>> assert no_max_move.iloc[-1] > 20

    :Example: state management
    --------------------------
    >>> old = a.iloc[:5000]
    >>> new = a.iloc[5000:]
    >>> old_ts = ewmrms_(old, 10)
    >>> new_ts = ewmrms(new, 10, **old_ts) # instantiation with previous ewma
    >>> ts = ewmrms(a,10)
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: Support for time & clock
    ------------------------------------
    >>> daily = a
    >>> monthly = daily.resample('M').last()
    >>> m_ts = ewmrms(monthly, 3) ## 3-month ewma run on monthly data
    >>> d_ts = ewmrms(daily, 3, 'm') ## 3-month ewma run on daily data
    >>> daily_resampled_to_month = d_ts.resample('M').last()
    >>> assert abs(daily_resampled_to_month - m_ts).max() < 1e-10

    So you can run a 3-monthly ewma on daily, where within month, most recent value is used with the EOM history.
    
    :Example: Support for dict/list of arrays
    --------------------------------------
    >>> x = pd.Series(np.random.normal(0,1,1000), drange(-999)); y = pd.Series(np.random.normal(0,1,1000), drange(-999))
    >>> a = dict(x = x, y = y)    
    >>> assert eq(ewmrms(dict(x=x, y=y),10), dict(x=ewmrms(x,10), y=ewmrms(y,10)))
    >>> assert eq(ewmrms([x,y],10), [ewmrms(x,10), ewmrms(y,10)])
    
             
    :Returns:
    -------
    an array/timeseries of ewma
    
    """
    state = state or {}
    return first_(_ewmrmst(a, n, time = time, axis=axis, exc_zero = exc_zero, max_move = max_move, **state))

def ewmstd_(a, n, time = None, min_sample=0.25, bias = False, axis=0, data = None, exc_zero = False, max_move = 0, instate = None):
    """
    Equivalent to ewmstd but returns a state parameter for instantiation of later calculations.
    See ewmstd documentation for more details
    """
    state = instate or {}
    return _data_state(['data', 't', 't0', 't1', 't2', 'w2'],_ewmstdt(a, n, time = time, min_sample=min_sample, axis=axis, exc_zero = exc_zero, max_move = max_move, calculator = stdev_calculation_ewm, **state))

ewmstd_.output = ['data', 'state']

def ewmstd(a, n, time = None, min_sample=0.25, bias = False, axis=0, data = None, exc_zero = False, max_move = 0, state = None):
    """
    ewmstd is equivalent to a.ewm(n).std() but with...
    - supports np.ndarrays as well as timeseries
    - handles nan by skipping them
    - allows state-management
    - ability to supply a 'clock' to the calculation
    
    :Parameters:
    ------------
    a : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    
    exc_zero: bool
        if True, will skip zeros inside a and will assume these are a result of forward filling
    max_move: float
        if positive, will "clip" a[i] at current stdev estimation * max_move
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    
    :Example: matching pandas
    -------------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> ts = ewmstd(a,10); df = a.ewm(10).std()
    >>> assert abs(ts-df).max()<1e-10
    >>> ts = ewmstd(a,10, bias = True); df = a.ewm(10).std(bias = True)
    >>> assert abs(ts-df).max()<1e-10


    :Example: exc_zero
    ------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> a[dt(-10):] = 0
    >>> exclude = ewmstd(a, 10, exc_zero = True)
    >>> assert exclude .iloc[-1] == exclude.iloc[-10]
    >>> no_exclude = ewmstd(a, 10)
    >>> assert no_exclude.iloc[-1] < no_exclude.iloc[-10]

    :Example: max_move
    ------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)) ## a series with stdev of 1 approximately
    >>> a.iloc[-1] = 100. ## put a bad data point with a huge spike...
    >>> max_move = ewmstd(a, 10, max_move = 4)
    >>> assert max_move.iloc[-1] < 2
    >>> no_max_move = ewmstd(a, 10)
    >>> assert no_max_move.iloc[-1] > 20

    :Example: numpy arrays support
    ------------------------------
    >>> assert eq(ewmstd(a.values, 10), ewmstd(a,10).values)

    :Example: nan handling
    ----------------------
    >>> a[a.values<-0.1] = np.nan
    >>> ts = ewmstd(a,10, time = 'i'); df = a.ewm(10).std() # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10
    >>> ts = ewmstd(a,10, time = 'i', bias = True); df = a.ewm(10).std(bias = True) # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10
    
        
    :Example: state management
    --------------------------
    >>> old = a.iloc[:5000]
    >>> new = a.iloc[5000:]
    >>> old_ts = ewmstd_(old, 10)
    >>> new_ts = ewmstd(new, 10, **old_ts) # instantiation with previous ewma
    >>> ts = ewmstd(a,10)
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: Support for time & clock
    ------------------------------------
    >>> daily = a
    >>> monthly = daily.resample('M').last()
    >>> m_ts = ewmstd(monthly, 3) ## 3-month ewma run on monthly data
    >>> d_ts = ewmstd(daily, 3, 'm') ## 3-month ewma run on daily data
    >>> daily_resampled_to_month = d_ts.resample('M').last()
    >>> assert abs(daily_resampled_to_month - m_ts).max() < 1e-10

    So you can run a 3-monthly ewma on daily, where within month, most recent value is used with the EOM history.
    
    :Example: Support for dict/list of arrays
    --------------------------------------
    >>> x = pd.Series(np.random.normal(0,1,1000), drange(-999)); y = pd.Series(np.random.normal(0,1,1000), drange(-999))
    >>> a = dict(x = x, y = y)    
    >>> assert eq(ewmstd(dict(x=x, y=y),10), dict(x=ewmstd(x,10), y=ewmstd(y,10)))
    >>> assert eq(ewmstd([x,y],10), [ewmstd(x,10), ewmstd(y,10)])
    
             
    :Returns:
    -------
    an array/timeseries of ewma
    
    """
    state = state or {}
    return first_(_ewmstdt(a, n, time = time, min_sample=min_sample, bias = bias, axis=axis, exc_zero = exc_zero, max_move = max_move,  calculator = stdev_calculation_ewm, **state))


def ewmvar_(a, n, time = None, min_sample=0.25, bias = False, axis=0, exc_zero = False, data = None, instate = None):
    """
    Equivalent to ewmvar but returns a state parameter for instantiation of later calculations.
    See ewmvar documentation for more details
    """
    state = instate or {}
    max_move = 0
    return _data_state(['data', 't', 't0', 't1', 't2', 'w2'],_ewmstdt(a, n, time = time, min_sample=min_sample, axis=axis, exc_zero = exc_zero, max_move = max_move, calculator = variance_calculation_ewm, **state))

def ewmvar(a, n, time = None, min_sample=0.25, bias = False, axis=0, exc_zero = False, data = None, state = None):
    """
    ewmstd is equivalent to a.ewm(n).var() but with...
    - supports np.ndarrays as well as timeseries
    - handles nan by skipping them
    - allows state-management
    - ability to supply a 'clock' to the calculation
    
    :Parameters:
    ------------
    a : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    exc_zero: bool
        if True, will skip zeros inside a and will assume these are a result of forward filling
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    
    :Example: matching pandas
    -------------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> ts = ewmvar(a,10); df = a.ewm(10).var()
    >>> assert abs(ts-df).max()<1e-10
    >>> ts = ewmvar(a,10, bias = True); df = a.ewm(10).var(bias = True)
    >>> assert abs(ts-df).max()<1e-10


    :Example: numpy arrays support
    ------------------------------
    >>> assert eq(ewmvar(a.values, 10), ewmvar(a,10).values)

    :Example: nan handling
    ----------------------
    >>> a[a.values<-0.1] = np.nan
    >>> ts = ewmvar(a,10, time = 'i'); df = a.ewm(10).var() # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10
    >>> ts = ewmvar(a,10, time = 'i', bias = True); df = a.ewm(10).var(bias = True) # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10
    
        
    :Example: state management
    --------------------------
    >>> old = a.iloc[:5000]
    >>> new = a.iloc[5000:]
    >>> old_ts = ewmvar_(old, 10)
    >>> new_ts = ewmvar(new, 10, **old_ts) # instantiation with previous ewma
    >>> ts = ewmvar(a,10)
    >>> assert eq(new_ts, ts.iloc[5000:])

    :Example: Support for time & clock
    ------------------------------------
    >>> daily = a
    >>> monthly = daily.resample('M').last()
    >>> m_ts = ewmvar(monthly, 3) ## 3-month ewma run on monthly data
    >>> d_ts = ewmvar(daily, 3, 'm') ## 3-month ewma run on daily data
    >>> daily_resampled_to_month = d_ts.resample('M').last()
    >>> assert abs(daily_resampled_to_month - m_ts).max() < 1e-10

    So you can run a 3-monthly ewma on daily, where within month, most recent value is used with the EOM history.
    
    :Example: Support for dict/list of arrays
    --------------------------------------
    >>> x = pd.Series(np.random.normal(0,1,1000), drange(-999)); y = pd.Series(np.random.normal(0,1,1000), drange(-999))
    >>> a = dict(x = x, y = y)    
    >>> assert eq(ewmvar(dict(x=x, y=y),10), dict(x=ewmvar(x,10), y=ewmvar(y,10)))
    >>> assert eq(ewmvar([x,y],10), [ewmvar(x,10), ewmvar(y,10)])
    
             
    :Returns:
    -------
    an array/timeseries of ewma
    
    """
    state = state or {}
    max_move = 0
    return first_(_ewmstdt(a, n, time = time, min_sample=min_sample, bias = bias, axis=axis, exc_zero = exc_zero, max_move = max_move, calculator = variance_calculation_ewm, **state))


ewmstd_.output = ['data', 'state']


def ewmcor_(a, b, n, time = None, min_sample = 0.25, bias = True, axis = 0, data = None, instate = None):
    """
    Equivalent to ewmcor but returns a state parameter for instantiation of later calculations.
    See ewmcor documentation for more details
    """
    state = instate or {}    
    return _data_state(['data', 't', 't0', 'a1', 'a2', 'b1', 'b2', 'ab', 'w2'], _ewmcort(a = a, b = b, n = n, time = time, min_sample=min_sample, bias = bias, axis=axis, **state))

ewmcor_.output = ['data', 'state']


def ewmcor(a, b, n, time = None, min_sample = 0.25, bias = True, axis = 0, data = None, state = None):
    """
    calculates pair-wise correlation between a and b.
    
    :Parameters:
    ----------
    a : array/timeseries
    b : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    min_sample : floar, optional
        minimum weight of observations before we return a reading. The default is 0.25. This ensures that we don't get silly numbers due to small population.
    bias : book, optional
        vol estimation for a and b should really by unbiased. Nevertheless, we track pandas and set bias = True as a default.
    axis : int, optional
        axis of calculation. The default is 0.
    data : place holder, ignore, optional
        ignore. The default is None.
    state : dict, optional
        Output from a previous run of ewmcor_. The default is None.
        
    :Example: matching pandas
    -------------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> b = pd.Series(np.random.normal(0,1,9000), drange(-8999))
    >>> ts = ewmcor(a, b, n = 10); df = a.ewm(10).corr(b)
    >>> assert abs(ts-df).max()<1e-10

    :Example: numpy arrays support
    ------------------------------
    >>> assert eq(ewmcor(a.values, b.values, 10), ewmcor(a, b, 10).values)

    :Example: nan handling
    ----------------------
    >>> a[a.values<-0.1] = np.nan
    >>> ts = ewmcor(a, b, 10, time = 'i'); df = a.ewm(10).corr(b) # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10

    :Example: state management
    --------------------------
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> b = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> old_a = a.iloc[:5000]; old_b = b.iloc[:5000]
    >>> new_a = a.iloc[5000:]; new_b = b.iloc[5000:]
    >>> old_ts = ewmcor_(old_a, old_b, 10)
    >>> new_ts = ewmcor(new_a, new_b, 10, **old_ts) # instantiation with previous ewma
    >>> ts = ewmcor(a,b,10)
    >>> assert eq(new_ts, ts.iloc[5000:])
    
    """
    state = state or {}    
    return first_(_ewmcort(a, b = b, n = n, time = time, min_sample=min_sample, bias = bias, axis=axis, **state))


def ewmGLM_(a, b, n, time = None, min_sample = 0.25, bias = True, data = None, instate = None):
    """
    Equivalent to ewmGLM but returns a state parameter for instantiation of later calculations.
    See ewmGLM documentation for more details
    """
    state = instate or {}    
    return _data_state(['data', 't', 't0', 'a2', 'ab'], _ewmGLMt(a = a, b = b, n = n, time = time, min_sample=min_sample, **state))

ewmGLM_.output = ['data', 'state']


def ewmGLM(a, b, n, time = None, min_sample = 0.25, bias = True, data = None, state = None):
    """
    Calculates a General Linear Model fitting b to a.
    
    :Parameters:
    ----------
    a : a 2-d array/pd.DataFrame of values fitting b
    b : a 1-d array/pd.Series
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    min_sample : floar, optional
        minimum weight of observations before we return the fitting. The default is 0.25. This ensures that we don't get silly numbers due to small population.
    data : place holder, ignore, optional
        ignore. The default is None.
    state : dict, optional
        Output from a previous run of ewmGLM_. The default is None.
        
    :Theory:
    --------
    See https://en.wikipedia.org/wiki/Generalized_linear_model for full details.
    Briefly, we assume b is single column while a is multicolumn. 
    We minimize least square error (LSE) fitting:
    
    >>> b[i] =\sum_j m_j a_j[i]    
    >>> LSE(m) = \sum_i w_i (b[i] - \sum_j m_j * a_j[i])^2

    >>> dLSE/dm_k = 0  
    >>> <==>  \sum_i w_i (b[i] - \sum_j m_j * a_j[i]) a_k[i] = 0
    >>> <==>  E(b*a_k) = m_k E(a_k^2) + sum_{j<>k} m_k E(a_j a_k) 
    
    E is expectation under weights w. And we can rewrite it as:
    
    >>> a2 x m = ab ## matrix multiplication
    >>> a2[i,j] = E(a_i * a_j)
    >>> ab[j] = E(a_j * b)
    >>> m = a2.inverse x ab ## matrix multiplication    


    :Example: simple fit
    --------------------
    >>> from pyg import *
    >>> a = pd.DataFrame(np.random.normal(0,1,(10000,10)), drange(-9999))
    >>> true_m = np.random.normal(1,1,10)
    >>> noise = np.random.normal(0,1,10000)
    >>> b = (a * true_m).sum(axis = 1) + noise
    
    >>> fitted_m = ewmGLM(a, b, 50)    
        
    """
    state = state or {}    
    return first_(_ewmGLMt(a = a, b = b, n = n, time = time, min_sample=min_sample, **state))



def ewmLR_(a, b, n, time = None, min_sample = 0.25, bias = True, axis = 0, c = None, m = None, instate = None):
    """
    Equivalent to ewmcor but returns a state parameter for instantiation of later calculations.
    See ewmcor documentation for more details
    """
    state = instate or {}    
    return _data_state(['c', 'm', 't', 't0', 'a1', 'a2', 'b1', 'b2', 'ab', 'w2'], 
                       _ewmLRt(a = a, b = b, n = n, time = time, min_sample=min_sample, bias = bias, axis=axis, **state),
                       output = ['c', 'm'])

ewmLR_.output = ['c', 'm', 'state']

def ewmLR(a, b, n, time = None, min_sample = 0.25, bias = True, axis = 0, c = None, m = None, state = None):
    """
    calculates pair-wise linear regression between a and b.
    We have a and b for which we want to fit:
    
    >>> b_i = c + m a_i 
    >>> LSE(c,m) = \sum w_i (c + m a_i - b_i)^2
    >>> dLSE/dc  = 0  <==> \sum w_i  (c + m a_i - b_i) = 0    [1]
    >>> dLSE/dm  = 0 <==> \sum w_i  a_i (c + m a_i - b_i) = 0 [2]

    >>> c     + mE(a)    = E(b)     [1]
    >>> cE(a) + mE(a^2)  = E(ab)    [2]

    >>> cE(a) + mE(a)^2  = E(a)E(n) [1] * E(a) 
    >>> m(E(a^2) - E(a)^2) = E(ab) - E(a)E(b)
    >>> m = covar(a,b)/var(a)
    >>> c = E(b) - mE(a)

    
    Parameters
    ----------
    a : array/timeseries
    b : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    min_sample : floar, optional
        minimum weight of observations before we return a reading. The default is 0.25. This ensures that we don't get silly numbers due to small population.
    bias : book, optional
        vol estimation for a and b should really by unbiased. Nevertheless, we track pandas and set bias = True as a default.
    axis : int, optional
        axis of calculation. The default is 0.
    c,m : place holder, ignore, optional
        ignore. The default is None.
    state : dict, optional
        Output from a previous run of ewmcor_. The default is None.
    
    :Example: numpy arrays support
    ------------------------------
    >>> assert eq(ewmLR(a.values, b.values, 10), ewmLR(a, b, 10).values)

    :Example: nan handling
    ----------------------
    >>> a[a.values<-0.1] = np.nan
    >>> ts = ewmcor(a, b, 10, time = 'i'); df = a.ewm(10).corr(b) # note: pandas assumes, 'time' pass per index entry, even if value is nan
    >>> assert abs(ts-df).max()<1e-10

    :Example: state management
    --------------------------
    >>> from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> b = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> old_a = a.iloc[:5000]; old_b = b.iloc[:5000]
    >>> new_a = a.iloc[5000:]; new_b = b.iloc[5000:]
    >>> old_ts = ewmLR_(old_a, old_b, 10)
    >>> new_ts = ewmLR(new_a, new_b, 10, **old_ts) # instantiation with previous ewma
    >>> ts = ewmLR(a,b,10)
    >>> assert eq(new_ts.c, ts.c.iloc[5000:])
    >>> assert eq(new_ts.m, ts.m.iloc[5000:])
    
    
    :Example:
    ---------
    >>> from pyg import *
    >>> a0 = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> a1 = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> b = (a0 - a1) + pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> a = pd.concat([a0,a1], axis=1)
    >>> LR = ewmLR(a,b,50)
    >>> assert abs(LR.m.mean()[0]-1)<0.5
    >>> assert abs(LR.m.mean()[1]+1)<0.5
    """
    state = state or {}
    return _data_state(['c', 'm'], _ewmLRt(a = a, b = b, n = n, time = time, min_sample=min_sample, bias = bias, axis=axis, **state),
                       output = ['c', 'm'])

ewmLR.output = ['c', 'm']

def ewmskew_(a, n, time = None, bias = False, min_sample = 0.25, axis=0, data = None, instate = None):
    """
    Equivalent to ewmskew but returns a state parameter for instantiation of later calculations.
    See ewmskew documentation for more details
    """
    state = instate or {}
    return _data_state(['data', 't', 't0', 't1', 't2', 't3'],_ewmskewt(a, n, time = time, bias = bias, min_sample = min_sample, axis = axis, **state))

ewmskew_.output = ['data', 'state']

def ewmskew(a, n, time = None, bias = False, min_sample = 0.25, axis=0, data = None, state = None):
    """
    Equivalent to a.ewm(n).skew() but with...
    - supports np.ndarrays as well as timeseries
    - handles nan by skipping them
    - allows state-management
    - ability to supply a 'clock' to the calculation
    
    :Parameters:
    ------------
    a : array/timeseries
    n : int/fraction
        The number or days (or a ratio) to scale the history
    time : Calendar, 'b/d/y/m' or a timeseries of time (use clock(a) to see output)
        If time parameter is provided, we allow multiple observations per unit of time. i.e., converging to the last observation in time unit. 
            - if we have intraday data, and set time = 'd', then 
            - the ewm calculation on last observations per day is what is retained. 
            - the ewm calculation on each intraday observation is same as an ewm(past EOD + current intraday observation)
    
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 
    
    :Example: matching pandas
    -------------------------
    >>> import pandas as pd; import numpy as np; from pyg import *
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> x = a.ewm(10).skew()

    >>> old = a.iloc[:10]
    >>> new = a.iloc[10:]
    >>> for f in [ewma_, ewmstd_, ewmrms_, ewmskew_, ]:
    >>>    both = f(a, 3)
    >>>    o = f(old, 3)
    >>>    n = f(new, 3, **o)
    >>>    assert eq(o.data, both.data.iloc[:10]) 
    >>>    assert eq(n.data, both.data.iloc[10:]) 
    >>>    assert both - 'data' == n - 'data'

    >>> assert abs(a.ewm(10).mean() - ewma(a,10)).max() < 1e-14
    >>> assert abs(a.ewm(10).std() - ewmstd(a,10)).max() < 1e-14

    :Example: numpy arrays support
    -------------------------------
    >>> assert eq(ewma(a.values, 10), ewma(a,10).values)

    :Example: nan handling
    ----------------------
    while panadas ffill values, timeseries skips nans:

    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999))
    >>> a[a.values>0.1] = np.nan
    >>> ts = ewma(a,10)
    >>> assert eq(ts[~np.isnan(ts)], ewma(a[~np.isnan(a)], 10))
    
    :Example: initiating the ewma with past state
    ------------------------------------------------
    >>> old = np.random.normal(0,1,100)
    >>> new = np.random.normal(0,1,100)
    >>> old_ = ewma_(old, 10)
    >>> new_ = ewma(new, 10, t0 = old_ewma.t0, t1 = old_ewma.t1) # instantiation with previous ewma
    >>> new_2 = ewma(np.concatenate([old,new]), 10)[-100:]
    >>> assert eq(new_ewma, new_ewma2)

    :Example: Support for time & clock
    -----------------------------------
    >>> daily = pd.Series(np.random.normal(0,1,10000), drange(-9999)).cumsum()
    >>> monthly = daily.resample('M').last()
    >>> m = ewma(monthly, 3) ## 3-month ewma run on monthly data
    >>> d = ewma(daily, 3, 'm') ## 3-month ewma run on daily data
    >>> daily_resampled_to_month = d.resample('M').last()
    >>> assert abs(daily_resampled_to_month - m).max() < 1e-10

    So you can run a 3-monthly ewma on daily, where within month, most recent value is used with the EOM history.

    
    :Returns:
    -------
    an array/timeseries of ewma
    
    """
    state = state or {}
    return first_(_ewmskewt(a, n, time = time, bias = bias, min_sample=min_sample, axis=axis, **state))




from scipy.optimize import minimize
from pyg_timeseries._rolling import ffill
from pyg_base import is_num
import numpy as np

def residual(c):
    return (1-c**2)**0.5


def _variance(weights, c):
    return weights.dot(c.dot(weights.transpose()))

def _addem(weights):
    return 1.0 - np.sum(weights)

def _ensure_initial_weights_verify_bounds(weights, min_weights, max_weights):
    return weights



def _single_period_max_diversify(c, min_weights, max_weights, weights = None, tol = 0.001):
    """
    This implements a one-step optimization of weights. We handle nan in various columns

    see also 
    robcarver17/march2023_independent_bets.py
    https://gist.github.com/robcarver17/c350ca7c73018c175393e44ad839c061

    """
    bad = np.isnan(c[0])
    good = ~bad
    t = c.shape[0]
    n = sum(good[good])
    if n == 1:
        weights = np.zeros(t)
        weights[good] = 1/n
        return weights
    if weights is None:
        weights = np.zeros(t)
        weights[good] = 1/n
    
    ## now we project onto the good area
    weights_ = weights[good]
    min_weights_ = min_weights[good]
    max_weights_ = max_weights[good]
    c_ = c[good,good]

    weights = _ensure_initial_weights_verify_bounds(weights = weights_, 
                                                    min_weights = min_weights_, 
                                                    max_weights=max_weights_)
    bounds = list(zip(min_weights_, max_weights_))
    cdict = [{"type": "eq", "fun": _addem}]
    ans = minimize(
           _variance,
           weights_,
           (c_,),
           method="SLSQP",
           bounds=bounds,
           constraints=cdict,
           tol=tol)
    rtn = np.zeros(t)
    rtn[good] = ans['x']
    return rtn
        
    


def maxdiv(correlation, min_weights = None, max_weights = None, max_change = None, weights = None):
    """
    ### let us set up the assets
    
    >>> sp = pd.Series(np.random.normal(0,1, 10000), drange(-9999))
    >>> dax_c = 0.3; dax = add_(pd.Series(np.random.normal(0,1, 5000), drange(-4999)) * residual(dax_c),  dax_c * sp)
    >>> cac_c = 0.1; cac = add_(pd.Series(np.random.normal(0,1, 4000), drange(-3999)) * residual(cac_c),  cac_c * sp)
    >>> ty_c = -0.2; ty = add_(pd.Series(np.random.normal(0,1, 8000), drange(-7999)) * residual(ty_c),  ty_c * sp)

    >>> dax[np.random.normal(0,1,len(dax)) > 3] = np.nan
    >>> cac[np.random.normal(0,1,len(cac)) > 3] = np.nan
    >>> ty[np.random.normal(0,1,len(ty)) > 3] = np.nan
    
    >>> rtn = df_concat([sp,dax,cac,ty], ['sp', 'dax', 'cac', 'ty'])
    >>> correlation = ewmcorr(rtn, 255)
    >>> assert correlation.shape == (10000,4,4)

    correlation[np.isnan(rtn.dax).values]
    correlation = ffill(correlation)
    """
    correlation = ffill(correlation)
    c = correlation[0]
    t = c.shape[0]
    if min_weights is None:
        min_weights = 0.
    if max_weights is None:
        max_weights = 1.
    if is_num(min_weights):
        min_weights = np.array([min_weights] * t)
    if is_num(max_weights):
        max_weights = np.array([max_weights] * t)
    if max_change is None:
        max_change = 1.
    if is_num(max_change):
        max_change = np.array([max_change] * t)




from pyg_timeseries import compiled
from pyg_base import getargspec


def test_ts_compiled():
    def function(a, b):
        return a + b
    f = compiled(function)
    assert f.fullargspec == getargspec(f)
    assert f(1,2) == 3



import numpy as np
from pyg_base import is_ts, is_pd

def least_squares(A, b, cost = None, **kwargs):
    """
    single value least-squares, 
    Optimizing ||Ax - b|| in L2 Norm

    :Parameters:
    ----------
    A: 1d or 2d np.array
        nans in A's are converted to zeros
        
    b : 1d np.array
        nans in b yields to removed rows of both A and b
        
    Returns
    -------
    np.array 1d
        optimal solution

    :Example:
    ---------
    >>> m = 20
    >>> n = 15
    >>> t = 10
    >>> np.random.seed(1)
    
    >>> A = np.random.randn(t, m, n)
    >>> b = np.random.randn(t, m)
    
    # Add noise...
    
    >>> b[np.random.randn(t, m)>1] = np.nan
    >>> A[np.random.randn(t, m, n)>1] = np.nan

    >>> dfA = pd.DataFrame(A.reshape(t, m * n), drange(1-t))
    >>> dfb = pd.DataFrame(b, drange(1-t))

    # solve...

    >>> solution = ts_iterate(functor = least_squares, variables = dict(A = dfA, b = dfb), dates = dfA)

    # merge with previous solution:
    
    >>> solution2 = ts_iterate(functor = least_squares, variables = dict(A = dfA, b = dfb), dates = dfA, data = pd_solution.drop(solution.index[::3]))
    >>> assert eq(solution2, solution)
    
    """
    import cvxpy as cp
    if cost is None:
        cost = cp.sum_squares
    m = b.shape[0]
    n = A.shape[0] // m if len(A.shape) == 1 else A.shape[1]
    if is_ts(A) or is_ts(b):
        raise ValueError('cannot get a timeseries')
    if is_pd(A):
        A = A.values
    if is_pd(b):
        b = b.values
    A = A.reshape(m,n) if len(A.shape) == 1 else A
    mask = ~np.isnan(b)
    b = b[mask]
    A = A[mask]
    A[np.isnan(A)] = 0
    x = cp.Variable(n)
    costs = cost(A @ x - b)
    prob = cp.Problem(cp.Minimize(costs))
    prob.solve()
    return x.value




import numpy as np
from pyg_base import calendar
from pyg_timeseries._rolling import shift, diff
from pyg_timeseries._ewm import ewma
import pandas as pd

ln2 = np.log(2)


def bar_daily_to_eod(bar, o = 'open', v = 'volume', cal = None, day_start = None, day_end = None):
    """
    Constructs an eod-bar from intraday data bar ASSUMING the bar provided at each intraday is actually a daily intraday bar... 
    i.e. every bar is actually the "day so far" bar.
    - hence, a day changes when the open changes
    - or when the volume drops
    
    if neither is provided, uses simple date to decide transition.
    
    ASSUMPTION: conversion to actual date is done using the calendar.trade_date function.
    
    Parameters
    ----------
    bar : pd.DataFrame
        bar timeseries
    o : str, optional
        name of the open column, default 'open'.
    v : str, optional
        name of the volume colume. The default is 'volume'.

    >>> from pyg import *; from pysys import *
    >>> bar = pd_read_npy(data_path('bc/ESA Index/2023/H/intraday.npy')); v = 'volume'; o = 'open'; cal = None
    >>> bar_eod(bar)

    >>> bar = pd_read_npy(data_path('bc/TYA Comdty/2023/H/intraday.npy')); v = 'volume'; o = 'open'; cal = None
    >>> bar_eod(bar)
    >>> o = 'open'; v = 'volume'

    """
    if len(bar) == 0:
        return bar
    elif len(bar) == 1:
        eod = bar
    else: #remove duplicates based on open/oi, always keep latest point
        if not v and not o:
            mask = diff(bar.index.day.values.astype(float),-1)!=0
        elif v and o:
            vmask = diff(bar[v],-1) > 0
            omask = diff(bar[o],-1)!=0
            omask.iloc[-1] = False
            mask = np.maximum(omask, vmask)        
        elif o:
            omask = diff(bar[o],-1)!=0
            omask.iloc[-1] = False
            mask = omask
        elif v:
            mask = vmask = diff(bar[v],-1) > 0  
        mask.iloc[-1] = True
        eod = bar[mask]
    cal = calendar(cal)    
    eod.index = [cal.trade_date(d, adj = 'f', day_start = day_start, day_end = day_end) for d in eod.index] 
    date_repeated = np.concatenate([eod.index[:-1] == eod.index[1:], np.array([False])])
    eod = eod[~date_repeated]
    last_date = cal.trade_date(bar.index[-1], adj = 'f', day_start = day_start, day_end = day_end)
    if last_date > eod.index[-1]:
        last_bar = bar.iloc[-1:]
        last_bar.index = [last_date]
        eod = pd.concat([eod, last_bar])
    return eod


def ohlc_r2(o1, h1, l1, c1, c0, model = 'yz', alpha = None, short_range = None, long_range = None):
    """

    Calculates a point-in-time R^2 estimation for variance from a bar.
    
    Yang Zhang original paper is here: http://www.atmif.com/papers/range.pdf
    Rogers et al, defining V_rs can be found here: http://www.skokholm.co.uk/wp-content/uploads/2016/01/RS_Yoon0.pdf (survey) 
    and here https://www.skokholm.co.uk/wp-content/uploads/2016/01/R_Satchell_HLOC.pdf
    
    We follow Yang Zhang notations but be aware of implementation differences:

    1) we choose not to remove drift... i.e.     
        In YZ paper, estimate for variance is given by     
        V_c = mean((c - mean(c))**2)
        We ignore the drift correction and use
        V_c = mean(c**2)

    2) We do not calculate the variance estimate, we calculate the daily R^2
    The process of calculating the average is left for later as (unbiased) mean of daily R^2 over a period.
    This is useful as user may want to 
        - "stitch together" the R^2 across multiple time series (e.g. future contracts) prior to variance calculation
        - calculate an ewma average rather than a linear average in the paper
        
    3) The distribution assumptions of all the fancy estimators may be false as there may be jumps, auto correlation etc. 
    The "reference" vol estimation is actually the close-close so you can choose to calibrate to it.
    Calibration matches short-range vol estimation over a rolling long range. 
    We find that calibration 
        - brings the estimators together in the mean.
        - hardly changing the turnover of the slow yz/rs/gk estimators.

    :Parameters:
    ============    
    o1,h1,l1,c1: timseries
        open/high/low/close today

    c0: timeseries
        close previous day
    
    model: str 
        type of r2 estimator required:
        'cc': close-close estimator
        'p' : Parkinson estimator 
        'rs': Roger Satchel estimator 
        'gk': Garman Klass estimator
        'yz': Yang Zhang (default) 
    
    alpha: float
        Only applied to yz estimator, determines ratio between o-c variance and o-h-c & o-l-c variance estimators
    
    short_range, long_range: int
        If supplied, will be used to calibrate estimator so that long-term vol estimation is close to the c-c estimator
        This is to compensate for distribution assumptions made by gk, yz and rs estimators.
        We find this makes the estimators match the mean better while hardly impacting turnover    

    :Example:
    =========
    >>> from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000), [date for date in drange(-2500, 0, '1h') if date.hour>5 and date.hour<22][:10000]) ## 16 observations per day each N(0,1), so daily vol is 4
    >>> price = rtn.cumsum()
    >>> price.plot()
    >>> daily = price.resample('d')
    >>> daily.count()
    >>> bar = df_concat([daily.first(), daily.max(), daily.min(), daily.last()], ['open', 'high', 'low', 'close'])
    >>> yz_raw = bar_r2(bar)

    >>> long_range = 100
    >>> yz = bar_r2(bar, short_range = 40, long_range = long_range)
    >>> cc = bar_r2(bar, model = 'cc', short_range = 20, long_range = long_range)
    >>> rs = bar_r2(bar, model = 'rs', short_range = 20, long_range = long_range)
    >>> gk = bar_r2(bar, model = 'gk', short_range = 20, long_range = long_range)
    >>> yz_raw_vol = ewma(yz_raw, 20) ** 0.5
    >>> yz_vol = ewma(yz, 20) ** 0.5
    >>> cc_vol = ewma(cc, 20) ** 0.5
    >>> rs_vol = ewma(rs, 20) ** 0.5
    >>> gk_vol = ewma(gk, 20) ** 0.5

    >>> print('\nmean vol was generated to be near 4 but yz with no calibration underestimates vol:\n', dict(yz_raw_mean = yz_raw_vol.mean(), yz_mean = yz_vol.mean(), cc_mean = cc_vol.mean()))
    >>> print('\nturnover of yz and yz_raw should be similar:\n', dict(cc_tover = tover(cc_vol), rs_tover = tover(rs_vol), gk_tover = tover(gk_vol), yz_tover = tover(yz_vol), yz_raw_tover = tover(yz_raw_vol)))
    
    >>> df_concat([cc_vol, yz_vol, yz_raw_vol, rs_vol, gk_vol], ['cc', 'yz', 'yz_raw', 'rs', 'gk']).plot()    

    """
    o = o1 - c0
    c = c1 - o1    
    cc = c1 - c0 # = o + c
    u = h1 - o1
    d = l1 - o1
    hl = h1 - l1
    
    
    V_cc = cc ** 2
    V_rs = u * (u - c) + d * (d - c) # Eq 3 # This equation seriously underestimates daily variance in practice
    alpha = alpha or 1.34 ## eq 10 and following discussion
    k = (alpha - 1) / (alpha + 1) # eq 10, note that for our cases, n+1/n-1 == 1 approx
    
    V_o = o ** 2  # Eq 5 
    V_c = c ** 2  # Eq 6
    V_p = (u-d) ** 2 / (4 * ln2)  # Eq 2
    V_gk = V_o - 0.383 * V_c + 1.364 * V_p + 0.019 * V_rs # Eq 4, Garman Klass estimator
    V_yz = V_o + k * V_c + (1-k) * V_rs # Eq 9 

    ### some cleaning choice
    V_yz[np.isnan(V_yz)] = V_gk ## if we have some nan, default to Garman Klass variance
    V_yz[np.isnan(V_yz)] = V_cc ## if still nan, use close-close variance
    V_yz[hl <= 0] = V_cc ## if we have bad high-low data, revert to close-close variance
    res = dict(yz = V_yz, gk = V_gk, p = V_p, c = V_c, cc = V_cc, rs = V_rs, o = V_o)[model.lower()]
    if long_range and short_range and model.lower()!='cc':
        cc_vol_avg = ewma(ewma(V_cc, short_range) ** 0.5, long_range)
        res_vol_avg = ewma(ewma(res, short_range) ** 0.5, long_range)
        ratio = (cc_vol_avg/res_vol_avg)**2
        res = ratio * res
    return res



def bar_r2(bar, o = 'open', h = 'high', l = 'low', c = 'close', model = 'yz', alpha = None, short_range = None, long_range = None):
    """

    Calculates a point-in-time R^2 estimation for variance from a bar.
    
    Yang Zhang original paper is here: http://www.atmif.com/papers/range.pdf
    Rogers et al, defining V_rs can be found here: http://www.skokholm.co.uk/wp-content/uploads/2016/01/RS_Yoon0.pdf (survey) 
    and here https://www.skokholm.co.uk/wp-content/uploads/2016/01/R_Satchell_HLOC.pdf
    
    We follow Yang Zhang notations but be aware of implementation differences:

    1) we choose not to remove drift... i.e.     
        In YZ paper, estimate for variance is given by     
        V_c = mean((c - mean(c))**2)
        We ignore the drift correction and use
        V_c = mean(c**2)

    2) We do not calculate the variance estimate, we calculate the daily R^2
    The process of calculating the average is left for later as (unbiased) mean of daily R^2 over a period.
    This is useful as user may want to 
        - "stitch together" the R^2 across multiple time series (e.g. future contracts) prior to variance calculation
        - calculate an ewma average rather than a linear average in the paper
        
    3) The distribution assumptions of all the fancy estimators may be false as there may be jumps, auto correlation etc. 
    The "reference" vol estimation is actually the close-close so you can choose to calibrate to it.
    Calibration matches short-range vol estimation over a rolling long range. 
    We find that calibration 
        - brings the estimators together in the mean.
        - hardly changing the turnover of the slow yz/rs/gk estimators.

    :Parameters:
    ============
    bar: pd.DataFrame
        bar data
    
    o,h,l,c: str
        column names for ohlc
    
    model: str 
        type of r2 estimator required:
        'cc': close-close estimator
        'p' : Parkinson estimator 
        'rs': Roger Satchel estimator 
        'gk': Garman Klass estimator
        'yz': Yang Zhang (default) 
    
    alpha: float
        Only applied to yz estimator, determines ratio between o-c variance and o-h-c & o-l-c variance estimators
    
    short_range, long_range: int
        If supplied, will be used to calibrate estimator so that long-term vol estimation is close to the c-c estimator
        This is to compensate for distribution assumptions made by gk, yz and rs estimators.
        We find this makes the estimators match the mean better while hardly impacting turnover    

    :Example:
    =========
    >>> from pyg import * 
    >>> rtn = pd.Series(np.random.normal(0,1,10000), [date for date in drange(-2500, 0, '1h') if date.hour>5 and date.hour<22][:10000]) ## 16 observations per day each N(0,1), so daily vol is 4
    >>> price = rtn.cumsum()
    >>> price.plot()
    >>> daily = price.resample('d')
    >>> daily.count()
    >>> bar = df_concat([daily.first(), daily.max(), daily.min(), daily.last()], ['open', 'high', 'low', 'close'])
    >>> yz_raw = bar_r2(bar)

    >>> long_range = 100
    >>> yz = bar_r2(bar, short_range = 40, long_range = long_range)
    >>> cc = bar_r2(bar, model = 'cc', short_range = 20, long_range = long_range)
    >>> rs = bar_r2(bar, model = 'rs', short_range = 20, long_range = long_range)
    >>> gk = bar_r2(bar, model = 'gk', short_range = 20, long_range = long_range)
    >>> yz_raw_vol = ewma(yz_raw, 20) ** 0.5
    >>> yz_vol = ewma(yz, 20) ** 0.5
    >>> cc_vol = ewma(cc, 20) ** 0.5
    >>> rs_vol = ewma(rs, 20) ** 0.5
    >>> gk_vol = ewma(gk, 20) ** 0.5

    >>> print('\nmean vol was generated to be near 4 but yz with no calibration underestimates vol:\n', dict(yz_raw_mean = yz_raw_vol.mean(), yz_mean = yz_vol.mean(), cc_mean = cc_vol.mean()))
    >>> print('\nturnover of yz and yz_raw should be similar:\n', dict(cc_tover = tover(cc_vol), rs_tover = tover(rs_vol), gk_tover = tover(gk_vol), yz_tover = tover(yz_vol), yz_raw_tover = tover(yz_raw_vol)))
    
    >>> df_concat([cc_vol, yz_vol, yz_raw_vol, rs_vol, gk_vol], ['cc', 'yz', 'yz_raw', 'rs', 'gk']).plot()    

    """
    o1 = bar[o]
    h1 = bar[h]
    l1 = bar[l]
    c1 = bar[c]
    c0 = shift(c1)
    return ohlc_r2(o1 = o1, h1 = h1, l1 = l1, c1 = c1, c0 = c0, model = model, alpha = alpha, short_range = short_range, long_range=long_range)
    



from pyg_base import is_int, pd2np
from pyg_timeseries._median import rolling_median
from pyg_timeseries._decorators import compiled, _data_state, first_
from pyg_timeseries._ewm import _w
import numpy as np

@pd2np
@compiled
def _zmooth(a, smooth, n, max_move, t0 = 0, t2 = 0, prev = np.nan, exc_zero = False):
    p = w = _w(n)
    res = np.empty_like(a)
    vol = 0 if t0 == 0 else np.sqrt(t2/t0)
    if max_move == 0:
        raise ValueError('must provide a positive max_move')
    for i in range(a.shape[0]):
        if np.isnan(a[i]):
            res[i] = np.nan
        else:
            if np.isnan(prev): ## prev value is nan, no reference to calculate "change" from
                res[i] = a[i]
            else:
                v = a[i] - prev
                sign = np.sign(v)
                if vol > 0 and abs(v) > max_move * vol:
                    if np.isnan(smooth[i]):
                        v = sign * max_move * vol 
                    elif np.sign(smooth[i] - prev) == sign:
                        v = smooth[i] - prev
                    else:
                        v = 0
                res[i] = prev + v
                ### now we update vol
                if exc_zero and v == 0:
                    pass
                else:
                    t0 = t0 * p + (1-w)
                    t2 = t2 * p + (1-w) * v**2
                    vol = 0 if t0 == 0 else np.sqrt(t2/t0)
            prev = res[i]
    return res, t0, t2, prev


def zmooth_(a, n, smooth = None, max_move = 4.2, exc_zero = False, instate = None):
    """
    A z-filter + smooth with state management
    See zmooth documentation for more details
    """
    state = instate or {}
    if smooth is None:
        smooth = a + np.nan
    elif is_int(smooth):
        smooth = rolling_median(a, smooth)
    return _data_state(['data', 't0', 't2', 'prev'], _zmooth(a = a, n = n, smooth = smooth, exc_zero = exc_zero, max_move = max_move,**state))

zmooth_.output = ['data', 'state']


def zmooth(a, n, smooth = None, max_move = 4.2, exc_zero = False, state = None):
    """
    This function combines a standard z-filter with a smooth
    smoothing smoothes data but also introduces a delay which we don't usually want
    
    So what we want to do is as follows:

    if a[i] - res[i-1] exceeds a threshold, 
        res[i] = smooth[i] provided it is in the direction of a[i] else res[i-1] + threshold[i]
    else:        
        res[i] = a[i]
    
    Finally:
    threshold[i] is defined as a multiple of rolling rms of ewmrms(a[i]-a[i-1],n)
    
    Note: 
    The usual z-filter is when smooth is not provided. 
    The problem with a z-filter is that it can get stuck near the old value if the jump is many standard deviations away
    

    >>> from pyg import * 
    >>> from pyg_timeseries import * 
    >>> rtn = np.random.normal(0,1,10000)
    >>> a = cumsum(rtn)
    
    ### now we add some noise..
    >>> mask5 = np.random.normal(0,1,10000)>1.5 
    >>> a[mask5] = (a + np.random.normal(0,5,10000))[mask5]
    >>> mask10 = np.random.normal(0,1,10000)>2    
    >>> a[mask10] = (a + np.random.normal(0,10,10000))[mask10]
    >>> smooth = rolling_median(a, 10)
    >>> z = zmooth(a = a, smooth = smooth, n = 18)
    >>> assert ts_cor(diff(z), diff(a)) > 0.5
    >>> assert ts_cor(diff(smooth), diff(a)) < 0.3    

    ## another way is by looking at when the maximum correlation occurs
    >>> shifted_cor = [ts_cor(shift(a,i), smooth) for i in range(10)]
    >>> assert shifted_cor.index(max(shifted_cor)) > 0 ## there is a delay introduced...
    >>> shifted_cor = [ts_cor(shift(a,i), z) for i in range(10)]
    >>> assert shifted_cor.index(max(shifted_cor)) == 0

    :Example: using prev_value + 4 stdev as next value
    ---------    
    >>> z = zmooth(a = a, smooth = None, n = 18, max_move = 2)
    >>> assert ts_cor(diff(z), diff(a)) > 0.8
    >>> shifted_cor = [ts_cor(shift(a,i), z) for i in range(10)]
    >>> assert shifted_cor.index(max(shifted_cor)) == 0
   
    :Example:
    --------
    >>> n = 10; max_move = 4
    """    
    state = state or {}
    if smooth is None:
        smooth = a + np.nan
    elif is_int(smooth):
        smooth = rolling_median(a, smooth)
    return first_(_zmooth(a = a, n = n, smooth = smooth, exc_zero = exc_zero, max_move = max_move, **state))




from pyg_base import eq, nona
from pyg_timeseries import fnna, ffill, na2v, v2na
from pyg_timeseries._math import stdev_calculation_ewm, variance_calculation_ewm, cor_calculation_ewm, skew_calculation
import numpy as np; import pandas as pd
from numpy import nan, array

def test_fnna():
    a = np.array([np.nan, np.nan, 1, np.nan, np.nan, 2, np.nan, np.nan])    
    assert fnna(a) == 2
    assert fnna(a,2) == 5
    assert fnna(a,-1) == 5
    assert fnna(a,-2) == 2
    assert fnna(a,3) is None

def test_na2v():
    a = np.array([np.nan, np.nan, 1, np.nan, np.nan, 2, np.nan, np.nan])    
    assert eq(na2v(a), np.array([0., 0., 1., 0., 0., 2., 0., 0.]))
    assert eq(na2v(a, 10), np.array([10., 10., 1., 10., 10., 2., 10., 10.]))

def test_v2na():
    a = np.array([np.nan, np.nan, 1, np.nan, np.nan, 2, np.nan, np.nan])    
    assert eq(v2na(a, 1), np.array([nan, nan, nan, nan, nan,  2., nan, nan]))
    assert eq(v2na(na2v(a)), a)

def test_nona():
    a = np.array([np.nan, np.nan, 1, np.nan, np.nan, 2, np.nan, np.nan])    
    assert eq(nona(a), array([1., 2.]))

    a2 = np.array([np.nan, 1 , 1, np.nan, 1, np.nan, np.nan, np.nan])    
    b = np.array([a,a2]).T
    
    assert len(nona(b)) == 4
    assert eq(nona(b), array([[nan,1,nan,2], [1,1,1,nan]]).T)


def test_negative_variance():
    assert np.isnan(stdev_calculation_ewm(10, 100000, 20, 0, 0))
    assert np.isnan(variance_calculation_ewm(10, 100000, 20, 0, 0))
    assert np.isnan(skew_calculation(10, 100000, 20, 20, True, 0))

    assert variance_calculation_ewm(10, 10, 100, 0, 1, bias = True) == 9
    assert variance_calculation_ewm(10, 10, 100, 0, 1, bias = False) == 9



from pyg_base import *
from pyg_base import drange, sub_, eq
from pyg_timeseries import ts_mean, ts_sum, ts_skew, ts_rms, ts_max, ts_min, ts_count, ts_median, cumsum, shift, diff, cumprod, ffill, bfill
from pyg_timeseries import expanding_max, expanding_mean, expanding_median, expanding_min, expanding_rank, expanding_rms, expanding_skew, expanding_std, expanding_sum
from pyg_timeseries import  rolling_max, rolling_mean, rolling_median, rolling_min, rolling_quantile, rolling_rank, rolling_rms, rolling_skew, rolling_std, rolling_sum
from pyg_timeseries import ewma, ewmrms, ewmstd, ewmskew, ewmvar
import numpy as np
import pandas as pd
from functools import partial
epsilon = 1e-8

def _nona_data():
    np.random.seed(0)
    s = pd.Series(np.random.normal(0,1,1000), drange(-999))
    df = pd.DataFrame(np.random.normal(0,1,(1000, 20)), drange(-999))
    return [s,df]

def _nan_data():
    np.random.seed(0)
    s = pd.Series(np.random.normal(0,1,1000), drange(-999))
    s[s<0.1] = np.nan
    df = pd.DataFrame(np.random.normal(0,1,(1000, 20)), drange(-999))
    df[df<0.1] = np.nan
    return [s,df]


def test_ts_pandas_match_with_ints():
    a = pd.Series([1,2,3], drange(2))
    assert eq(diff(a), a.diff())

def test_ts_pandas_match():
    n2p = dict(ts_mean = lambda a, **kwargs: a.mean(**kwargs),
            ts_sum = lambda a, **kwargs: a.sum(**kwargs),
            ts_skew = lambda a, **kwargs: a.skew(**kwargs),
            ts_rms = lambda a, **kwargs: (a**2).mean(**kwargs)**0.5,
            ts_max = lambda a, **kwargs: a.max(**kwargs),
            ts_min = lambda a, **kwargs: a.min(**kwargs),
            ts_count = lambda a, **kwargs: a.count(**kwargs), 
            ts_median = lambda a, **kwargs: a.median(**kwargs))

    n2f = dict(ts_mean = ts_mean ,
            ts_sum = ts_sum,
            ts_skew = ts_skew,
            ts_rms = ts_rms,
            ts_max = ts_max,
            ts_min = ts_min,
            ts_count = ts_count, 
            ts_median = ts_median)
               
    s, df = _nona_data()
    for n, p in n2p.items():
        f = n2f[n]
        assert abs(f(s) - p(s)) < epsilon
        assert abs(f(df) - p(df)).max() < epsilon

    
    s, df = _nan_data()
    for n, p in n2p.items():
        f = n2f[n]
        assert abs(f(s) - p(s)) < epsilon
        assert abs(f(df) - p(df)).max() < epsilon
        assert abs(f(df, axis = 1) - p(df, axis = 1)).max() < epsilon
        

def test_expanding_pandas_match():
    n2p = dict(expanding_max = lambda a, axis = 0, **kwargs: a.expanding(axis = axis).max(**kwargs),
            expanding_mean = lambda a, axis = 0, **kwargs: a.expanding(axis = axis).mean(**kwargs),
            expanding_median = lambda a, axis = 0, **kwargs: a.expanding(axis = axis).median(**kwargs),
            expanding_min = lambda a, axis = 0, **kwargs: a.expanding(axis = axis).min(**kwargs),
            expanding_rms = lambda a, axis = 0, **kwargs: (a**2).expanding(axis = axis).mean(**kwargs) **0.5,
            expanding_std = lambda a, axis = 0, **kwargs: a.expanding(axis = axis).std(**kwargs),
            expanding_sum = lambda a, axis = 0, **kwargs: a.expanding(axis = axis).sum(**kwargs))



    n2f = dict(expanding_max = expanding_max, 
               expanding_mean = expanding_mean, 
               expanding_median = expanding_median,
               expanding_min = expanding_min,
               expanding_rank = expanding_rank, 
               expanding_rms = expanding_rms,
               expanding_skew = expanding_skew, 
               expanding_std = expanding_std,
               expanding_sum= expanding_sum)
               
    s, df = _nona_data()
    for n, p in n2p.items():
        f = n2f[n]
        assert abs(f(s) - p(s)).max() < epsilon
        assert abs(f(df) - p(df)).max().max() < epsilon

    s, df = _nan_data()
    for n, p in n2p.items():
        f = n2f[n]
        assert abs(f(s) - p(s)).max() < epsilon
        assert abs(f(df) - p(df)).max().max() < epsilon
        assert abs(f(df, axis = 1) - p(df, axis = 1)).max().max() < epsilon
        
def test_rolling_pandas_match():
    n2f = dict(
        rolling_max = rolling_max, 
        rolling_mean = rolling_mean, 
        rolling_median = rolling_median, 
        rolling_min = rolling_min, 
        rolling_quantile = partial(rolling_quantile, quantile = 0.3), 
        rolling_rank = rolling_rank, 
        rolling_rms = rolling_rms, 
        rolling_skew = rolling_skew, 
        rolling_std = rolling_std, 
        rolling_sum = rolling_sum)
    
    n2p = dict(
        rolling_max = lambda a, n, axis = 0, **kwargs: a.rolling(n, axis = axis).max(**kwargs), 
        rolling_mean = lambda a, n, axis = 0, **kwargs: a.rolling(n, axis = axis).mean(**kwargs), 
        rolling_median = lambda a, n, axis = 0, **kwargs: a.rolling(n, axis = axis).median(**kwargs), 
        rolling_min = lambda a, n, axis = 0, **kwargs: a.rolling(n, axis = axis).min(**kwargs), 
        rolling_quantile = lambda a, n, q = 0.3, axis = 0, **kwargs: a.rolling(n, axis = axis).quantile(q, **kwargs), 
        rolling_rms = lambda a, n, axis = 0, **kwargs: (a**2).rolling(n, axis = axis).mean(**kwargs)**0.5, 
        rolling_std = lambda a, n, axis = 0, **kwargs: a.rolling(n, axis = axis).std(**kwargs), 
        rolling_sum = lambda a, n, axis = 0, **kwargs: a.rolling(n, axis = axis).sum(**kwargs))
    
    s, df = _nona_data()
    for n, p in n2p.items():
        f = n2f[n]
        # args = (f(s, 10),p(s, 10)); kwargs = {}; self = sub_
        assert abs(sub_(f(s, 10),p(s, 10))).max() < epsilon
        assert abs(f(df,10) - p(df,10)).max().max() < epsilon
        assert abs(f(df,10,axis=1) - p(df,10, axis = 1)).max().max() < epsilon
        
def test_ewma_pandas_match():
    n2f = dict(ewma = ewma, ewmstd = ewmstd, ewmrms = ewmrms, ewmskew = ewmskew, ewmvar = ewmvar)
    n2p = dict(ewma = lambda a, n, axis = 0: a.ewm(n, axis = axis).mean(),
               ewmstd = lambda a, n, axis = 0: a.ewm(n, axis = axis).std(),
               ewmvar = lambda a, n, axis = 0: a.ewm(n, axis = axis).var(),
               ewmrms = lambda a, n, axis = 0: (a**2).ewm(n, axis = axis).mean()**0.5)
    s, df = _nona_data()

    for n, p in n2p.items():
        f = n2f[n]
        assert abs(sub_(f(s, 10), p(s, 10))).max() < epsilon
        assert abs(f(df,10) - p(df,10)).max().max() < epsilon
        assert abs(f(df,10,axis=1) - p(df,10, axis = 1)).max().max() < epsilon
        
    # for data with nan, pandas ewma can be calculated as clocking time whenever there is nan, hence time = 'index' match.
    # however, this is not the case for std(). I really have no idea how panda handles nans
    s, df = _nan_data()
    n = 'ewma'
    p = n2p[n]
    f = n2f[n]
    assert abs(sub_(f(s, 10, time = 'i'),p(s, 10))).max() < epsilon
    assert abs(f(df,10, time = 'i') - p(df,10)).max().max() < epsilon
    assert abs(f(df,10, time = 'i', axis=1) - p(df,10, axis = 1)).max().max() < epsilon


def test_bits_vs_pandas():
    s, df = _nona_data()    
    assert abs(cumsum(s) - s.cumsum()).max() <1e-13
    assert abs(cumsum(df) - df.cumsum()).max().max() < 1e-13

    for n in [-4,-1,0,1,10]:
        assert abs(shift(s,n) - s.shift(n)).max() <1e-13
        assert abs(shift(df,n) - df.shift(n)).max().max() < 1e-13
        assert abs(diff(s,n) - s.diff(n)).max() <1e-13
        assert abs(diff(df,n) - df.diff(n)).max().max() < 1e-13

    s, df = _nan_data()    
    assert abs(np.log10(cumprod(s)) - np.log10(s).cumsum()).max()<1e-13
    assert abs(np.log10(cumprod(df)) - np.log10(df).cumsum()).max().max()<epsilon

    assert eq(ffill(s), s.fillna(method = 'ffill'))






from pyg_base import is_ts, loop
import pandas as pd

@loop(list, dict, tuple)
def _latest_value(value, date, default = None):
    if is_ts(value):
        v = value.loc[:date]
        if len(v):
            return v.iloc[-1].values 
        else:
            return default
    else:
        return value

def _latest_date_value(value, date, default = None):
    if is_ts(value):
        v = value.loc[:date]
        if len(v):    
            return v.index[-1], v.iloc[-1].values 
        else:
            return None, default
    else:
        return None, value


def ts_iterate(functor, variables, dates, data = None):
    """
    This is a generic functor iterating over timeseries.
    The output of previous calculation is a pd.DataFrame
    """
    if is_ts(dates):
        dates = dates.index
    ds = dates.drop(sorted(set(dates) & set(data.index))) if data is not None else dates
    previous_date = None
    values = []
    data_ = None
    maximal_date = data.index[-1] if is_ts(data) and len(data) else None
    for date in ds:
        variables_ = _latest_value(variables, date)
        if maximal_date and maximal_date > date:  ## we are back-filling here
            latest_date, latest_data = _latest_date_value(data, date)
            if latest_data is not None and latest_date is not None:
                if previous_date is None or latest_date > previous_date:
                    data_ = latest_data
        data_ = functor(data = data_, date = date, **variables_)
        previous_date  = date
        values.append(data_)
    result = pd.DataFrame(values, ds)
    if is_ts(data):
        result = pd.concat([data, result]).sort_index()
    return result       




import numpy as np
from pyg_timeseries._rolling import _fnna, _vec
from pyg_timeseries._decorators import _data_state, first_
from pyg_base import pd2np, as_list, loop_all, loop, is_pd


__all__ = ['rolling_quantile', 'rolling_quantile_']

def _as_strided(a, L, S = 1):
    nrows = ((a.size - L)//S) + 1
    n = a.strides[0]
    return np.lib.stride_tricks.as_strided(a, shape = (nrows, L), strides = (S*n,n))

def _cast_strided_result(a, va, res, mask, n):
    n0 = int(_fnna(a, n))
    if n>0:
        mask[:n0] = False
    else:
        mask[n0+1:] = False
    va[mask] = res
    va[~mask] = np.nan
    return va

@loop_all
@pd2np
def _rolling_quantile(a, n, quantile, vec = None):
    vec = _vec(a, vec, 0)
    if len(vec):
        a_ = np.concatenate([vec,a])
    else:
        a_ = a
    mask = ~np.isnan(a_)
    na = a_[mask]
    if len(na) < n:
        return a + np.nan
    quantile = np.array(quantile)
    va = np.empty_like(a_) if len(quantile.shape) == 0 else np.empty([a_.shape[0], quantile.shape[0]])
    strided = _as_strided(na, abs(n), 1)
    res = np.quantile(strided, quantile, axis = 1).T
    rtn = _cast_strided_result(a_, va, res, mask, n)
    if len(vec):
        rtn = rtn[-len(a):]
    return rtn, na[-(n-1):]

def rolling_quantile(a, n, quantile = 0.5, axis = 0, data = None, state = None):
    """
    equivalent to a.rolling(n).quantile(q) except...
    - supports numpy arrays
    - supports multiple q values

    :Example:
    -------
    >>> from pyg import *; import pandas as pd; import numpy as np
    >>> a = pd.Series(np.random.normal(0,1,10000), drange(-9999)); n = 100; quantile = 0.3; state = None
    >>> res = rolling_quantile(a, 100, 0.3)
    >>> assert sub_(res, a.rolling(100).quantile(0.3)).max() < 1e-13

    :Example: multiple quantiles
    ---------------------------------------------
    >>> res = rolling_quantile(a, 100, [0.3, 0.5, 0.75])
    >>> assert abs(res[0.3] - a.rolling(100).quantile(0.3)).max() < 1e-13

    :Example: state management
    ---------------------------------------------
    >>> res = rolling_quantile(a, 100, 0.3)
    >>> old = rolling_quantile_(a.iloc[:2000], 100, 0.3)
    >>> new = rolling_quantile(a.iloc[2000:], 100, 0.3, **old)
    >>> both = pd.concat([old.data, new])
    >>> assert eq(both, res)
    
    :Parameters:
    ----------------
    a : array/timeseries
    n : integer
        window size.
    q : float or list of floats in [0,1]
        quantile(s).
    data: None.
        unused at the moment. Allow code such as func(live, **func_(history)) to work
    state: dict, optional
        state parameters used to instantiate the internal calculations, based on history prior to 'a' provided. 

    :Returns:
    -------
    timeseries/array of quantile(s)

    """
    qs = as_list(quantile)
    if len(getattr(a, 'shape', [])) == 2 and a.shape[1] > 1:
        if len(qs) > 1:
            raise ValueError('Can do multiple quantiles %s only for single-column data'%qs)
        else:
            qs = qs[0]
    state = state or {}
    res = first_(_rolling_quantile(a, n = n , quantile = qs, axis = axis, **state))
    qs = as_list(quantile)
    @loop(list, dict)
    def add_qs(res):
        if is_pd(res) and len(res.shape) == 2 and res.shape[1] == len(qs):
            res.columns = qs
        return res
    return add_qs(res)

def rolling_quantile_(a, n, quantile = 0.5, axis = 0, data = None, instate = None):
    """
    Equivalent to rolling_quantile(a) but returns also the state. 
    For full documentation, look at rolling_quantile.__doc__    
    """
    qs = as_list(quantile)
    if len(getattr(a, 'shape', [])) == 2 and a.shape[1] > 1:
        if len(qs) > 1:
            raise ValueError('Can do multiple quantiles %s only for single-column data'%qs)
        else:
            qs = qs[0]
    state = instate  or {}
    res = _data_state(['data','vec'],_rolling_quantile(a, n = n, quantile = qs, axis = axis, **state))
    qs = as_list(quantile)
    @loop(list, dict)
    def add_qs(res):
        if is_pd(res) and len(res.shape) == 2 and res.shape[1] == len(qs):
            res.columns = qs
        return res
    return add_qs(res)

rolling_quantile_.ouput = ['data', 'state']


import numpy as np
import pandas as pd
from pyg_base import eq, nona, Dict, drange, alphabet
from pyg_timeseries import ffill, bfill
from pyg_timeseries import ewma, ewmskew, ewmrms, ewmstd, ewmvar, rolling_sum, rolling_mean, rolling_rms, rolling_std, rolling_skew, rolling_max, rolling_min, rolling_median
from pyg_timeseries import expanding_mean, expanding_sum, expanding_std, expanding_rms, cumsum, cumprod
from pyg_timeseries import ts_mean, ts_count, ts_max, ts_std, ts_rms, ts_skew, ts_median


np.random.seed(seed=0)

def _data():
    index = drange(-99)
    a2 = np.random.normal(0,1, (100,26))
    a1 = np.random.normal(0,1,100)
    na2 = np.random.normal(0,1, (100,26))
    na2[na2<0.1] = np.nan
    na1 = np.random.normal(0,1,100)
    na1[na1<0.1] = np.nan
    df = pd.DataFrame(a2, index, list(alphabet))
    ndf = pd.DataFrame(na2, drange(-99), list(alphabet))
    s = pd.Series(a1, index)
    ns = pd.Series(na1, index)
    return Dict(a1 = a1, a2 = a2, na1 = na1, na2 = na2, df = df, ndf = ndf, s = s, ns = ns, index = index)

d = _data()

def test_ts_vs_numpy():
    for f in ewma, ewmskew, ewmrms, ewmstd, ewmvar, rolling_sum, rolling_mean, rolling_rms, rolling_std, rolling_skew, rolling_max, rolling_min, rolling_median:
        assert eq(f(d.ndf, 10).values, f(d.ndf.values, 10))
        assert eq(f(d.df, 10).values, f(d.df.values, 10))
        assert eq(f(d.s, 10).values, f(d.s.values, 10))
        assert eq(f(d.ns, 10).values, f(d.ns.values, 10))
        assert eq(f(d.ndf, 5, axis=1).values, f(d.ndf.values, 5, axis=1))
        assert eq(f(d.df, 5, axis=1).values, f(d.df.values, 5, axis=1))

    for f in expanding_mean, expanding_sum, expanding_std, expanding_rms, cumsum, cumprod, ffill, bfill:
        assert eq(f(d.ndf).values, f(d.ndf.values))
        assert eq(f(d.df).values, f(d.df.values))
        assert eq(f(d.s).values, f(d.s.values))
        assert eq(f(d.ns).values, f(d.ns.values))
        assert eq(f(d.ndf, axis=1).values, f(d.ndf.values, axis=1))
        assert eq(f(d.df, axis=1).values, f(d.df.values, axis=1))

    for f in ts_mean, ts_count, ts_max, ts_std, ts_rms, ts_skew, ts_median:
        assert eq(f(d.ndf).values, f(d.ndf.values))
        assert eq(f(d.df).values, f(d.df.values))
        assert eq(f(d.ndf, axis=1).values, f(d.ndf.values, axis=1))
        assert eq(f(d.df, axis=1).values, f(d.df.values, axis=1))
        assert eq(f(d.s), f(d.s.values))
        assert eq(f(d.ns), f(d.ns.values))

    
def test_ts_and_nans():
    """
    We show that we can take an array, remove nans, run it without the nans and then resample back, and we get the same as running it outright
    """
    for f in ewma, ewmskew, ewmrms, ewmstd, ewmvar, rolling_sum, rolling_mean, rolling_rms, rolling_std, rolling_skew, rolling_max, rolling_min, rolling_median:
        assert eq(f(d.ns, 10), f(nona(d.ns), 10).reindex(d.ns.index))
        assert eq(f(d.ndf, 10)['c'], f(nona(d.ndf['c']), 10).reindex(d.ndf.index))

    for f in expanding_mean, expanding_sum, expanding_std, expanding_rms, cumsum, cumprod:
        assert eq(f(d.ns), f(nona(d.ns)).reindex(d.ns.index))
        assert eq(f(d.ndf)['c'], f(nona(d.ndf['c'])).reindex(d.ndf.index))
    
    for f in ts_mean, ts_count, ts_max, ts_std, ts_rms, ts_skew, ts_median:
        assert eq(f(d.ns), f(nona(d.ns)))
        assert eq(f(d.ndf)['c'], f(nona(d.ndf['c'])))
    

def test_ts_rolling_mean_with_time():
    a = np.arange(10) * 1.
    n = 2
    time = np.array([1,1,2,2,3,3,4,4,5,5.])
    raw = rolling_mean(a, n)
    timed = 2*rolling_mean(a, n, time = time)
    assert eq(raw, np.array([np.nan, 0.5, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5]))
    assert eq(timed, np.array([np.nan, np.nan, 1+2, 1+3 , 3+4, 3+5. , 5+6, 5+7 , 7+8, 7+9]))
    timed = rolling_sum(a, n, time = time)
    assert eq(timed, np.array([np.nan, np.nan, 1+2, 1+3 , 3+4, 3+5. , 5+6, 5+7 , 7+8, 7+9]))
    timed = rolling_std(a, n, time = time) ** 2
    assert np.max(np.abs(timed - np.array([np.nan, np.nan, 0.5, 2. , 0.5, 2. , 0.5, 2. , 0.5, 2. ]))[2:]) < 1e-10
    timed = 2 * (rolling_rms(a, n, time = time) ** 2)
    assert np.max(np.abs((timed - np.array([ np.nan,  np.nan,   1+4,  1+9.,  9+16.,  9+25.,  25+36.,  25+49., 49+64., 49+81.]))[2:])) < 1e-10



from pyg_timeseries._decorators import compiled, mask_nans, apply_along_first_axis
from pyg_timeseries._ewm import ewma, ewma_, ewmstd, ewmvar, ewmstd_, ewmrms, ewmrms_, ewmskew, ewmskew_, ewmcor, ewmcor_, ewmcorr, ewmcorr_, ewmcorrelation, ewmcorrelation_, ewmcovar, ewmcovar_, ewmvar_, ewmLR, ewmLR_, ewmGLM, ewmGLM_
from pyg_timeseries._min import rolling_min, rolling_min_, expanding_min, expanding_min_
from pyg_timeseries._max import rolling_max, rolling_max_, expanding_max, expanding_max_
from pyg_timeseries._median import rolling_median, rolling_median_, expanding_median
from pyg_timeseries._rank import rolling_rank, rolling_rank_, expanding_rank
from pyg_timeseries._rolling import ffill, ffill_, ewfill, bfill, fnna, fnnz, diff, shift, ratio, rolling_mean, rolling_sum, rolling_rms, rolling_std, rolling_skew, \
           diff_, shift_, ratio_, rolling_mean_, rolling_sum_, rolling_rms_, rolling_std_, rolling_skew_, v2na, na2v, buffer_, buffer, rolling_tover
from pyg_timeseries._stride import rolling_quantile, rolling_quantile_
from pyg_timeseries._expanding import cumsum, cumprod, cumsum_, cumprod_, \
                expanding_mean, expanding_sum, expanding_rms, expanding_std, expanding_skew, \
                expanding_mean_, expanding_sum_, expanding_rms_, expanding_std_, expanding_skew_
from pyg_timeseries._ts import ts_std, ts_sum, ts_mean, ts_skew, ts_count, ts_min, ts_max, ts_rms, ts_median, ts_cor, \
                                            ts_interval, ts_std_, ts_sum_, ts_mean_, ts_skew_, ts_count_, ts_min_, ts_max_, ts_rms_, ts_cor_
                                            
from pyg_timeseries._ewmxo import ou_factor, ewmacd, ewmacd_, ewmvol_, ewmvol
from pyg_timeseries._xrank import xrank
from pyg_timeseries._ewmcombine import ewmcombine, ewmcombined
from pyg_timeseries._pandas import fnna_like, reindex_3d
from pyg_timeseries._opt import least_squares
from pyg_timeseries._iter import ts_iterate
#from pyg_timeseries._linalg import eigenvalues, eigenvectors, det
from pyg_timeseries._matmul import matmul
from pyg_timeseries._bar import bar_r2, bar_daily_to_eod, ohlc_r2
from pyg_timeseries._track import minimize_tracking_error
from pyg_timeseries._multibuffer import multibuffer, near_correlation_matrix, beta_correlation_matrix
from pyg_timeseries._zmooth import zmooth, zmooth_



from pyg_base import pd2np, is_num, is_rng, is_pd, loop, is_arr, df_reindex
from pyg_timeseries._rolling import fnna, na2v
import pandas as pd
import numpy as np

@loop(list)
@pd2np
def _fnna_like(source, target = 1., default = np.nan):
    """
    takes a target value, list or an array and match source shape. Then matches the initial nan pattern

    Parameters
    ----------
    source : dataframe
        the source to copy
    target : value/list/array, optional
        The source we want to fill with nan when source is nan
    default : float/int, optional
        The value before nan

    Returns
    -------
    a : array matching shape of source

    :Example:
    --------
    
    >>> source = np.array([[np.nan, 1.], [1.,1.]])

    >>> fnna_like(source, 2)
    >>> array([[nan,  2.],
    >>>        [ 2.,  2.]])

    >>> fnna_like(source, [4,5])
    >>> array([[nan,  5.],
    >>>        [ 4.,  5.]])

    >>> import pandas as pd
    >>> source = pd.DataFrame(np.array([[np.nan, np.nan, 1.], [np.nan,2., 3.], [np.nan, np.nan, np.nan, ]]), [1,2,3], ['a', 'b', 'c'])

    >>> fnna_like(source, 1)
    >>>     a    b    c
    >>> 1 NaN  NaN  1.0
    >>> 2 NaN  1.0  1.0
    >>> 3 NaN  1.0  1.0

    """
    f = fnna(source)
    if is_num(target) or is_rng(target) or (is_arr(target) and len(target.shape) == 1 and target.shape[0] == source.shape[1]):
        a = np.full(source.shape, fill_value = target)
    else:
        a = target.copy()
    a = a.astype(type(default))
    if len(source.shape) == 1:
        a[:f] = default
    else:
        for i, j in enumerate(f):
            a[:j,i] = default
    return a

def fnna_like(source, target = 1., default = np.nan):
    if is_pd(source) and is_pd(target):
        target = target.reindex(source.index)
    return _fnna_like(source, target, default)


def reindex_3d(array, index, original_index, method = None, limit = None):
    if len(array.shape)<3:
        if isinstance(array, np.ndarray):
            ts = pd.Series(array, index = original_index) if len(array.shape) == 1 else pd.DataFrame(array, index = original_index)
        else:
            ts = array
        return df_reindex(ts, index = index, method = method, limit = limit)                    
    default = np.full(array[0].shape, method if is_num(method) else np.nan)
    idx = df_reindex(pd.Series(range(array.shape[0]), original_index), index, method = None if is_num(method) else method, limit = limit).values
    res = np.array([default if np.isnan(idx[i]) else array[int(idx[i])] for i in range(len(idx))])
    if is_num(method):
        res = na2v(res, method)
    return res
    
